#!/usr/bin/env python
#
# orthoeditprofile.py - The OrthoEditProfile class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoEditProfile` class, an interaction
:class:`.Profile` for :class:`.OrthoPanel` views.
"""

import logging

import numpy                        as np

import fsl.utils.idle               as idle
import fsl.utils.deprecated         as deprecated
import fsleyes_props                as props
import fsleyes.actions              as actions
import fsleyes.actions.copyoverlay  as copyoverlay
import fsleyes.editor               as fsleditor
import fsleyes.gl.routines          as glroutines
import fsleyes.gl.annotations       as annotations
from . import                          orthoviewprofile


log = logging.getLogger(__name__)


class OrthoEditProfile(orthoviewprofile.OrthoViewProfile):
    """The ``OrthoEditProfile`` class is an interaction profile for use with
    the :class:`.OrthoPanel` class. It gives the user the ability to make
    changes to :class:`.Image` overlays, by using the functionality of the
    :mod:`~fsleyes.editor` package.


    **Modes**


    The ``OrthoEditProfile`` has the following modes, in addition to those
    already defined by the :class:`.OrthoViewProfile`:

    =========== ===============================================================
    ``sel``     Select mode. The user is able to manually add voxels to the
                selection using a *cursor*. The cursor size can be changed
                with the :attr:`selectionSize` property, and the cursor can be
                toggled between a 2D square and a 3D cube via the
                :attr:`selectionIs3D` property. If the :attr:`drawMode`
                property is ``True``, selected voxels are immediately filled
                with the :attr:`fillValue` when the mouse is released.

    ``desel``   Deselect mode. Identical to ``sel`` mode, except that the
                cursor is used to remove voxels from the selection. If the
                :attr:`drawMode` property is ``True``, selected voxels are
                immediately set to 0 when the mouse is released.

    ``chsize``  Change-size mode. The use can change the :attr:`selectionSize`
                attribute via the mouse wheel.

    ``selint``  Select by intensity mode. The user can select a voxel, and
                grow the selected region based on its intensity.

    ``fill``    Fill mode. The user can click on a voxel and set its
                selected state, and the state of adjacent voxels. Restricted
                to 2D (see :attr:`selectionIs3D`).

    ``chthres`` Change-threshold mode. The user can change the
                :attr:`intensityThres` via the mouse wheel.

    ``chrad``   Change-radius mode. The user can change the
                :attr:`searchRadius` via the mouse wheel.
    =========== ===============================================================


    **Actions**


    The ``OrthoEditProfile`` defines the following actions, on top of those
    already defined by the :class:`.OrthoViewProfile`:

    .. autosummary::
       :nosignatures:

       undo
       redo
       clearSelection
       fillSelection
       invertSelection
       eraseSelection
       copySelection
       pasteSelection


    **Annotations**


    The ``OrthoEditProfile`` class uses :mod:`.annotations` on the
    :class:`.SliceCanvas` panels, displayed in the :class:`.OrthoPanel`,
    to display information to the user. Two annotations are used:

     - The *cursor* annotation. This is a :class:`.Rect` annotation
       representing a cursor at the voxel, or voxels, underneath the
       current mouse location.

     - The *selection* annotation. This is a :class:`.VoxelSelection`
       annotation which displays the :class:`.Selection`.


    **The display space**


    The ``OrthoEditProfile`` class has been written in a way which requires
    the :class:`.Image` instance that is being edited to be displayed in
    *scaled voxel* (a.k.a. ``pixdim``) space.  The :class:`.OrthoEditToolBar`
    uses a :class:`.DisplaySpaceWarning` widget to warn the user if the
    :attr:`.DisplayContext.displaySpace` is not set appropriately.
    """


    selectionCursorColour = props.Colour(default=(1, 1, 0, 0.7))
    """Colour used for the cursor annotation. """


    selectionOverlayColour = props.Colour(default=(1, 0.25, 1, 0.4))
    """Colour used for the selection annotation, which displays the voxels
    that are currently selected.
    """


    locationFollowsMouse = props.Boolean(deafult=True)
    """If ``True``, when the user is drawing/erasing/selectiong by clicking and
    dragging with the mouse, the :attr:`.DisplayContext.location` is updated to
    track the mouse.

    Users running on a slower machine may wish to disable this option.
    """


    showSelection = props.Boolean(default=True)
    """When :attr:`drawMode` is ``False,` the selection overlay can be hidden
    by setting this to ``False``.
    """


    selectionSize = props.Int(minval=1, maxval=100, default=3, clamped=True)
    """In ``sel`` and ``desel`` modes, defines the size of the selection
    cursor.
    """


    selectionIs3D = props.Boolean(default=False)
    """In ``sel`` and ``desel`` mode, toggles the cursor between a 2D square
    and a 3D cube. In ``selint`` mode, toggles the selection space between the
    current slice, and the full 3D volume.
    """


    fillValue = props.Real(default=1, clamped=True)
    """The value used when drawing/filling voxel values - all voxels in the
    selection will be filled with this value.
    """


    eraseValue = props.Real(default=0, clamped=True)
    """The value used when erasing voxel values - all voxels in the
    selection will be filled with this value.
    """


    drawMode = props.Boolean(default=True)
    """If ``True``, when in ``sel`` or ``desel`` mode, clicks and click+
    drags cause the image to be immediately modified. Otherwise, editing
    is a two stage process (as described in the :class:`.Editor` class
    documentation).

    This setting is enabled by default, because it causes FSLeyes to behave
    like FSLView. However, all advanced editing/selection capabilities are
    disabled when ``drawMode`` is ``True``.
    """


    intensityThres = props.Real(
        minval=0.0, maxval=1.0, default=0, clamped=False)
    """In ``selint`` mode, the maximum distance, in intensity, that a voxel
    can be from the seed location, in order for it to be selected.
    Passed as the ``precision`` argument to the
    :meth:`.Selection.selectByValue` method.
    """


    intensityThresLimit = props.Real(minval=0.0, default=0, clamped=True)
    """This setting controls the maximum value for the :attr:`itensityThres`
    property. It is set automatically from the data when an :class:`.Image`
    is first selected, but can also be manually controlled via this property.
    """


    localFill = props.Boolean(default=False)
    """In ``selint`` mode, if this property is ``True``, voxels can only be
    selected if they are adjacent to an already selected voxel. Passed as the
    ``local`` argument to the :meth:`.Selection.selectByValue` method.
    """


    limitToRadius = props.Boolean(default=False)
    """In ``selint`` mode, if this property is ``True``, the search region
    will be limited to a sphere (in the voxel coordinate system) with its
    radius specified by the :attr:`searchRadius` property.
    """


    searchRadius = props.Real(
        minval=0.01, maxval=200, default=0.0, clamped=False)
    """In ``selint`` mode, if :attr:`limitToRadius` is true, this property
    specifies the search sphere radius. Passed as the ``searchRadius``
    argument to the :meth:`.Selection.selectByValue` method.
    """


    targetImage = props.Choice()
    """By default, all modifications that the user makes will be made on the
    currently selected overlay (the :attr:`.DisplayContext.selectedOverlay`).
    However, this property  can be used to select a different image as the
    target for modifications.

    This proprty is mostly useful when in ``selint`` mode - the selection
    can be made based on the voxel intensities in the currently selected
    image, but the selection can be filled in another iamge (e.g. a
    mask/label image).

    This property is updated whenever the :class:`.OverlayList` or the
    currently selected overlay changes, so that it contains all other
    overlays which have the same dimensions as the selected overlay.
    """


    def __init__(self, viewPanel, overlayList, displayCtx):
        """Create an ``OrthoEditProfile``.

        :arg viewPanel:   The :class:`.OrthoPanel` instance.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        """

        # The currently selected overlay -
        # the overlay being edited.
        self.__currentOverlay = None

        # The 'clipboard' is created by
        # the copySelection method - it
        # contains a numpy array which
        # was copied from another overlay.
        # The clipboard source refers to
        # the overlay that the clipboard
        # was copied from.
        self.__clipboard       = None
        self.__clipboardSource = None

        # An Editor instance is created for each
        # Image overlay (on demand, as they are
        # selected), and kept in this dictionary
        # (which contains {Image : Editor} mappings).
        self.__editors = {}

        # Ref to each canvas on the ortho panel
        self.__xcanvas = viewPanel.getXCanvas()
        self.__ycanvas = viewPanel.getYCanvas()
        self.__zcanvas = viewPanel.getZCanvas()

        # The current selection is shown on each
        # canvas - a ref to the SelectionAnnotation
        # is kept here
        self.__xselAnnotation = None
        self.__yselAnnotation = None
        self.__zselAnnotation = None

        # When in draw/select/deselect/selint
        # modes, a Rect annotation is shown
        # on the canvases at the current mouse
        # location.
        self.__xcurAnnotation = None
        self.__ycurAnnotation = None
        self.__zcurAnnotation = None

        # A few performance optimisations are made
        # when in selint mode and limitToRadius is
        # active - the __record/__getSelectionMerger
        # methods populate these fields.
        self.__mergeMode   = None
        self.__mergeBlock  = None
        self.__merge3D     = None
        self.__mergeRadius = None

        # If the view panel performance is not
        # set to maximum, set the initial
        # locationFollowsMouse value to False
        perf = viewPanel.sceneOpts.performance
        self.locationFollowsMouse = perf == 3

        # These property values are cached
        # on a per-overlay basis. When an
        # overlay is re-selected, its values
        # are restored from the cache.
        self.__cache = props.PropCache(
            self,
            ['targetImage',
             'intensityThres',
             'intensityThresLimit',
             'searchRadius'],
            self.currentOverlay,
            [(overlayList, 'overlays'),
             (displayCtx,  'selectedOverlay')])

        orthoviewprofile.OrthoViewProfile.__init__(
            self,
            viewPanel,
            overlayList,
            displayCtx,
            ['sel', 'desel', 'chsize', 'selint',
             'fill', 'chthres', 'chrad'])

        self.mode = 'nav'

        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)

        self.addListener('targetImage',
                         self.name,
                         self.__targetImageChanged)
        self.addListener('mode',
                         self.name,
                         self.__modeChanged)
        self.addListener('drawMode',
                         self.name,
                         self.__drawModeChanged)
        self.addListener('selectionOverlayColour',
                         self.name,
                         self.__selectionColoursChanged)
        self.addListener('selectionCursorColour',
                         self.name,
                         self.__selectionColoursChanged)
        self.addListener('showSelection',
                         self.name,
                         self.__showSelectionChanged)
        self.addListener('intensityThresLimit',
                         self.name,
                         self.__selintThresLimitChanged)
        self.addListener('intensityThres',
                         self.name,
                         self.__selintPropertyChanged)
        self.addListener('searchRadius',
                         self.name,
                         self.__selintPropertyChanged)
        self.addListener('localFill',
                         self.name,
                         self.__selintPropertyChanged)
        self.addListener('limitToRadius',
                         self.name,
                         self.__selintPropertyChanged)

        self.__selectedOverlayChanged()
        self.__drawModeChanged()


    def destroy(self):
        """Removes some property listeners, destroys the :class:`.Editor`
        instances, and calls :meth:`.OrthoViewProfile.destroy`.
        """

        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.overlayList.removeListener('overlays',        self.name)

        for editor in self.__editors.values():
            editor.destroy()

        self.__destroyAnnotations()

        self.__editors         = None
        self.__xcanvas         = None
        self.__ycanvas         = None
        self.__zcanvas         = None
        self.__xselAnnotation  = None
        self.__yselAnnotation  = None
        self.__zselAnnotation  = None
        self.__xcurAnnotation  = None
        self.__ycurAnnotation  = None
        self.__zcurAnnotation  = None
        self.__currentOverlay  = None
        self.__clipboard       = None
        self.__clipboardSource = None
        self.__cache           = None

        orthoviewprofile.OrthoViewProfile.destroy(self)


    def deregister(self):
        """Destroys all :mod:`.annotations`, and calls
        :meth:`.OrthoViewProfile.deregister`.
        """

        self.__destroyAnnotations()
        orthoviewprofile.OrthoViewProfile.deregister(self)


    def __destroyAnnotations(self):
        """Called by other methods. Destroys the :class:`.SelectionAnnotation`
        and :class:`.Rect` cursor annotation objects, if they exist.
        """

        xannot = self.__xcanvas.getAnnotations()
        yannot = self.__ycanvas.getAnnotations()
        zannot = self.__zcanvas.getAnnotations()

        annots = [(self.__xselAnnotation, xannot),
                  (self.__xcurAnnotation, xannot),
                  (self.__yselAnnotation, yannot),
                  (self.__ycurAnnotation, yannot),
                  (self.__zselAnnotation, zannot),
                  (self.__zcurAnnotation, zannot)]

        for annotObj, annotMgr in annots:
            if annotObj is not None:
                annotMgr.dequeue(annotObj, hold=True)
                annotObj.destroy()

        self.__xselAnnotation = None
        self.__yselAnnotation = None
        self.__zselAnnotation = None
        self.__xcurAnnotation = None
        self.__ycurAnnotation = None
        self.__zcurAnnotation = None


    def currentOverlay(self):
        """Returns the overlay that is currently registered with this
        ``OrthoEditProfile``.
        """
        return self.__currentOverlay


    def editor(self, overlay):
        """Return the :class:`.Editor` associated with the given overlay.
        Raises a :exc:`KeyError` if there is no editor fo the overlay.
        """
        return self.__editors[overlay]


    @actions.action
    def createMask(self):
        """Create a 3D mask which has the same size as the currently selected
        overlay, and insert it into the overlay list.
        """
        if self.__currentOverlay is None:
            return

        overlay = self.__currentOverlay
        editor  = self.__editors[overlay]
        display = self.displayCtx.getDisplay(overlay)
        name    = '{}_mask'.format(display.name)
        data    = editor.getSelection().getSelection()
        data    = np.array(data, dtype=overlay.dtype)
        mask    = copyoverlay.copyImage(self.overlayList,
                                        self.displayCtx,
                                        self.__currentOverlay,
                                        createMask=True,
                                        copy4D=False,
                                        copyDisplay=False,
                                        name=name,
                                        data=data)
        self.displayCtx.selectOverlay(mask)


    @actions.action
    def clearSelection(self):
        """Clears the current selection. See :meth:`.Editor.clearSelection`.
        """

        if self.__currentOverlay is None:
            return

        editor = self.__editors[self.__currentOverlay]

        editor.getSelection().clearSelection()

        self.__refreshCanvases()


    @actions.action
    def fillSelection(self):
        """Fills the current selection with the :attr:`fillValue`. See
        :meth:`.Editor.fillSelection`.
        """
        if self.__currentOverlay is None:
            return

        editor = self.__editors[self.__currentOverlay]

        if self.targetImage is not None:
            editor = self.__getTargetImageEditor(editor)

        # TODO  You could get the bounded selection from
        # fillSelection, and pass it to clearSelection
        editor.startChangeGroup()
        editor.fillSelection(self.fillValue)
        editor.clearSelection()
        editor.endChangeGroup()


    @actions.action
    def invertSelection(self):
        """Inverts the current selection. See :meth:`.Editor.invertSelection`.
        """
        if self.__currentOverlay is None:
            return

        editor = self.__editors[self.__currentOverlay]

        if self.targetImage is not None:
            editor = self.__getTargetImageEditor(editor)

        editor.startChangeGroup()
        editor.invertSelection()
        editor.endChangeGroup()


    @actions.action
    def eraseSelection(self):
        """Fills the current selection with zero. See
        :meth:`.Editor.fillSelection`.
        """
        if self.__currentOverlay is None:
            return

        editor = self.__editors[self.__currentOverlay]

        if self.targetImage is not None:
            editor = self.__getTargetImageEditor(editor)

        # See TODO in fillSaelection
        editor.startChangeGroup()
        editor.fillSelection(self.eraseValue)
        editor.clearSelection()
        editor.endChangeGroup()


    @actions.action
    def copySelection(self):
        """Copies the data within the selection from the currently selected
        overlay, and stores it in an internal "clipboard".
        """

        overlay = self.__currentOverlay

        if overlay is None:
            return

        editor = self.__editors[overlay]

        self.__clipboard       = editor.copySelection()
        self.__clipboardSource = overlay

        self.__setCopyPasteState()


    @actions.action
    def pasteSelection(self):
        """Pastes the data currently stored in the clipboard into the currently
        selected image, if possible.
        """

        if self.__currentOverlay is None:
            return

        overlay         = self.__currentOverlay
        clipboard       = self.__clipboard
        clipboardSource = self.__clipboardSource

        if     overlay   is None:                  return
        if     clipboard is None:                  return
        if not clipboardSource.sameSpace(overlay): return

        editor = self.__editors[overlay]

        if self.targetImage is not None:
            editor = self.__getTargetImageEditor(editor)

        editor.startChangeGroup()
        editor.pasteSelection(clipboard)
        editor.clearSelection()
        editor.endChangeGroup()


    @actions.action
    def undo(self):
        """Un-does the most recent change to the selection or to the
        :class:`.Image` data. See :meth:`.Editor.undo`.
        """

        if self.__currentOverlay is None:
            return

        editor = self.__editors[self.__currentOverlay]

        # We're disabling notification of changes to the selection
        # during undo/redo. This is because a single undo
        # will probably involve multiple modifications to the
        # selection (as changes are grouped by the editor),
        # with each of those changes causing the selection object
        # to notify its listeners. As one of these listeners is a
        # SelectionTexture, these notifications can get expensive,
        # due to updates to the GL texture buffer. So we disable
        # notification, and then manually refresh the texture
        # afterwards
        with editor.getSelection().skipAll():
            change = editor.undo()

        if any([isinstance(c, fsleditor.SelectionChange) for c in change]):
            if not self.__xselAnnotation.destroyed:
                self.__xselAnnotation.texture.refresh()

        self.__refreshCanvases()


    @actions.action
    def redo(self):
        """Re-does the most recent undone change to the selection or to the
        :class:`.Image` data. See :meth:`.Editor.redo`.
        """

        if self.__currentOverlay is None:
            return

        editor = self.__editors[self.__currentOverlay]

        with editor.getSelection().skipAll():
            change = editor.redo()

        if any([isinstance(c, fsleditor.SelectionChange) for c in change]):
            if not self.__xselAnnotation.destroyed:
                self.__xselAnnotation.texture.refresh()

        self.__refreshCanvases()


    @deprecated.deprecated('0.31.0',
                           '1.0.0',
                           'Use fsleyes.editor.isEditable instead')
    def isEditable(self, overlay):
        """Returns ``True`` if the given overlay is editable, ``False``
        otherwise.
        """
        return fsleditor.isEditable(overlay, self.displayCtx)


    def __modeChanged(self, *a):
        """Called when the :attr:`.Profile.mode` changes. If the mode is
        changed to ``'fill'``, the :attr:`selectionIs3D` option is set to
        ``False``.
        """

        if self.mode == 'fill':
            self.selectionIs3D = False


    def __drawModeChanged(self, *a):
        """Called when the :attr:`drawMode` changes. Updates the enabled
        state of various actions that are irrelevant when in draw mode.
        """

        # The only possible profile modes when
        # drawMode==True are sel/desel/fill.
        if self.drawMode and self.mode not in ('nav', 'sel', 'desel', 'fill'):
            self.mode = 'sel'

        if self.drawMode: self.getProp('mode').disableChoice('selint', self)
        else:             self.getProp('mode').enableChoice( 'selint', self)

        self.clearSelection .enabled = not self.drawMode
        self.fillSelection  .enabled = not self.drawMode
        self.eraseSelection .enabled = not self.drawMode
        self.invertSelection.enabled = not self.drawMode

        with props.skip(self, 'showSelection', self.name):
            self.showSelection = True

        self.__updateTargetImage()
        self.__setCopyPasteState()

        if self.__currentOverlay is not None:
            editor = self.__editors[self.__currentOverlay]
            editor.getSelection().clearSelection()
            self.__refreshCanvases()


    def __setCopyPasteState(self):
        """Enables/disables the :meth:`copySelection`/ :meth:`pasteSelection`
        actions as needed.
        """

        overlay   = self.__currentOverlay
        clipboard = self.__clipboard
        source    = self.__clipboardSource

        enableCopy  = (not self.drawMode)     and \
                      (overlay is not None)

        enablePaste =  enableCopy             and \
                      (clipboard is not None) and \
                      (overlay.sameSpace(source))

        self.copySelection .enabled = enableCopy
        self.pasteSelection.enabled = enablePaste


    def __selectionColoursChanged(self, *a):
        """Called when either of the :attr:`selectionOverlayColour` or
        :attr:`selectionCursorColour` properties change.

        Updates the  :mod:`.annotations` colours accordingly.
        """
        if self.__xselAnnotation is None:
            return

        self.__xselAnnotation.colour = self.selectionOverlayColour
        self.__yselAnnotation.colour = self.selectionOverlayColour
        self.__zselAnnotation.colour = self.selectionOverlayColour


    def __showSelectionChanged(self, *a):
        """Called when the :attr:`showSelection` property changes. Shows/
        hides the :class:`.VoxelSelection` annotations accordingly.
        """
        if self.__xselAnnotation is None:
            return

        self.__xselAnnotation.enabled = self.showSelection
        self.__yselAnnotation.enabled = self.showSelection
        self.__zselAnnotation.enabled = self.showSelection
        self.__refreshCanvases()


    def __updateTargetImage(self):
        """Resets the value and choices on the :attr:`targetImage`.
        It is populated with all :class:`.Image` instances which are in the
        same space as the currently selected overlay.
        """

        with props.suppress(self, 'targetImage'):
            self.targetImage = None

        overlay = self.__currentOverlay

        if overlay is None:
            return

        compatibleOverlays = [None]
        if not self.drawMode:
            for ovl in self.overlayList:
                if ovl is not overlay                         and \
                   fsleditor.isEditable(ovl, self.displayCtx) and \
                   overlay.sameSpace(ovl):
                    compatibleOverlays.append(ovl)

        self.getProp('targetImage').setChoices(compatibleOverlays,
                                               instance=self)


    def __getTargetImageEditor(self, srcEditor):
        """If the :attr:`targetImage` is set to an image other than the
        currently selected one, this method returns an :class:`.Editor`
        for the target image.
        """

        if self.targetImage is None:
            return srcEditor

        tgtEditor = self.__editors[self.targetImage]
        srcSel    = srcEditor.getSelection()
        tgtSel    = tgtEditor.getSelection()

        tgtSel.setSelection(srcSel.getSelection(), (0, 0, 0))
        srcSel.clearSelection()

        return tgtEditor


    def __targetImageChanged(self, *a):
        """Called every time the :attr:`targetImage` is changed. Makes sure
        that an :class:`.Editor` instance for the selected target image exists.
        """

        image = self.targetImage

        if image is None: image = self.__currentOverlay
        if image is None: return

        editor = self.__editors.get(image, None)

        if editor is None:
            editor = fsleditor.Editor(image,
                                      self.overlayList,
                                      self.displayCtx)
            self.__editors[image] = editor


    def __setPropertyLimits(self):
        """Called by the :meth:`__selectedOverlayChanged` method.
        """

        overlay = self.__currentOverlay
        if overlay is None:
            # TODO Set to defaults? Probably
            #      not necessary
            return

        # If the image data is of an integer
        # type, we set limits on the fill/
        # erase values, so the user can't
        # enter an out-of-bounds value.
        if issubclass(overlay.dtype.type, np.integer):
            dmin = np.iinfo(overlay.dtype).min
            dmax = np.iinfo(overlay.dtype).max
        else:
            dmin = None
            dmax = None

        self.setAttribute('fillValue',  'minval', dmin)
        self.setAttribute('fillValue',  'maxval', dmax)
        self.setAttribute('eraseValue', 'minval', dmin)
        self.setAttribute('eraseValue', 'maxval', dmax)

        # Retrieve cached values. The cached
        # targetImage is set in the
        # __selecteDOverlayChanged method.
        thres  = self.__cache.get(overlay, 'intensityThres',      None)
        limit  = self.__cache.get(overlay, 'intensityThresLimit', None)
        radius = self.__cache.get(overlay, 'searchRadius',        None)

        # Set sensible initial values
        if limit is None or limit == 0:
            dmin, dmax = overlay.dataRange
            limit      = (dmax - dmin) / 2.0

        if thres is None: thres = 0
        else:             thres = min(thres, limit)

        # Set (what I think to be) a sensible
        # upper limit for the search radius.
        # This is just for slider widgets -
        # the seacrhRadius is unclamped, so
        # the user can manually enter any
        # value in spin boxes.
        dimSizes    = np.array(overlay.pixdim[:3]) * overlay.shape[:3]
        radiusLimit = max(dimSizes) / 4.0
        if radius is None or radius == 0:
            radius = radiusLimit / 4.0

        # Initialise property values, or
        # restore them from the cache.
        with props.skip(self, 'searchRadius', self.name):
            self.setAttribute('searchRadius', 'maxval', radiusLimit)
            self.searchRadius = radius

        with props.skip(self, 'intensityThres', self.name):
            self.setAttribute('intensityThres', 'maxval', limit)
            self.intensityThres = thres

        with props.skip(self, 'intensityThresLimit', self.name):
            self.intensityThresLimit = limit


    def __selectedOverlayChanged(self, *a):
        """Called when either the :class:`.OverlayList` or
        :attr:`.DisplayContext.selectedOverlay` change.

        Destroys all old :mod:`.annotations`. If the newly selected overlay is
        an :class:`Image`, new annotations are created.
        """
        # Overview:
        #  1. Destroy Editor instances associated with
        #     overlays that no longer exist
        #
        #  2. Destroy old canvas annotations
        #
        #  3. Remove property listeners on editor/selection
        #     objects associated with the previous overlay
        #
        #  4. Load/create a new Editor for the new overlay
        #
        #  5. Transfer the exsiting selection to the new
        #     overlay if possible.
        #
        #  6. Add property listeners to the editor/selection
        #
        #  7. Create canvas annotations
        #
        # Here we go....

        oldOverlay = self.__currentOverlay
        overlay    = self.displayCtx.getSelectedOverlay()

        # If the selected overlay hasn't changed,
        # we don't need to do anything
        if overlay == oldOverlay:
            self.__updateTargetImage()
            return

        # Destroy all existing canvas annotations
        self.__destroyAnnotations()

        # Remove property listeners from the
        # editor/selection instances associated
        # with the previously selected overlay
        if oldOverlay is not None:
            editor = self.__editors[oldOverlay]

            log.debug('De-registering listeners from Editor {} ({})'.format(
                id(editor), oldOverlay.name))
            self.undo.unbindProps('enabled', editor.undo)
            self.redo.unbindProps('enabled', editor.redo)

        # Make sure that the newly
        # selected overlay is editable
        if fsleditor.isEditable(overlay, self.displayCtx):
            self.__currentOverlay = overlay
        else:
            self.__currentOverlay = None

        # Update the limits/options
        # on all properties.
        self.__updateTargetImage()
        self.__setPropertyLimits()
        self.__setCopyPasteState()

        # If there is no selected overlay,
        # don't do anything more.
        if self.__currentOverlay is None:
            return

        # Update the limits/options on all properties.
        self.__setPropertyLimits()
        self.__setCopyPasteState()

        # Load the editor for the overlay (create
        # one if necessary), and add listeners to
        # some editor/selection properties
        editor = self.__editors.get(overlay, None)

        if editor is None:
            editor = fsleditor.Editor(overlay,
                                      self.overlayList,
                                      self.displayCtx)
            self.__editors[overlay] = editor

        # Transfer or clear the selection
        # for the old overlay.
        if oldOverlay is not None:

            oldSel = self.__editors[oldOverlay].getSelection()

            # Currently we only transfer
            # the selection for images
            # with the same dimensions/space
            if oldOverlay.sameSpace(overlay):

                log.debug('Transferring selection from {} to {}'.format(
                    oldOverlay.name,
                    overlay.name))

                newSel = editor.getSelection()
                newSel.setSelection(oldSel.getSelection(), (0, 0, 0))
            else:
                oldSel.clearSelection()

        # Restore the targetImage for this
        # overlay, if there is a cached value
        targetImage = self.__cache.get(overlay, 'targetImage', None)
        if targetImage not in self.overlayList:
            targetImage = None

        with props.skip(self, 'targetImage', self.name):
            self.targetImage = targetImage

        # Register property listeners with the
        # new Editor and Selection instances.
        log.debug('Registering listeners with Editor {} ({})'.format(
            id(editor),
            self.__currentOverlay.name))

        # Bind undo/redo action enabled states
        self.undo.bindProps('enabled', editor.undo)
        self.redo.bindProps('enabled', editor.redo)

        # Create a selection annotation and
        # a cursor annotation for each canvas
        sels         = []
        curs         = []
        cursorKwargs = {'colour'  : self.selectionCursorColour,
                        'width'   : 2,
                        'expiry'  : 0.5,
                        'enabled' : False}

        opts = self.displayCtx.getOpts(overlay)

        for c in [self.__xcanvas, self.__ycanvas, self.__zcanvas]:

            sels.append(annotations.VoxelSelection(
                c.getAnnotations(),
                editor.getSelection(),
                opts,
                colour=self.selectionOverlayColour))

            curs.append(annotations.Rect(
                c.getAnnotations(),
                (0, 0), 0, 0,
                **cursorKwargs))

        self.__xselAnnotation = sels[0]
        self.__yselAnnotation = sels[1]
        self.__zselAnnotation = sels[2]
        self.__xcurAnnotation = curs[0]
        self.__ycurAnnotation = curs[1]
        self.__zcurAnnotation = curs[2]

        xannot = self.__xcanvas.getAnnotations()
        yannot = self.__ycanvas.getAnnotations()
        zannot = self.__zcanvas.getAnnotations()

        xannot.obj(self.__xselAnnotation, hold=True)
        xannot.obj(self.__xcurAnnotation, hold=True)
        yannot.obj(self.__yselAnnotation, hold=True)
        yannot.obj(self.__ycurAnnotation, hold=True)
        zannot.obj(self.__zselAnnotation, hold=True)
        zannot.obj(self.__zcurAnnotation, hold=True)

        self.__refreshCanvases()


    def __getVoxelLocation(self, canvasPos):
        """Returns the voxel location, for the currently selected overlay,
        which corresponds to the specified canvas position. Returns ``None``
        if the current canvas position is out of bounds for the current
        overlay.
        """

        if self.__currentOverlay is None:
            return None

        opts = self.displayCtx.getOpts(self.__currentOverlay)
        return opts.getVoxel(canvasPos)


    def __hideCursorAnnotation(self):
        """Configures all of the :class:`.Rect` cursor annotations so that
        they will not be shown on the next canvas refresh.
        """
        xcur = self.__xcurAnnotation
        ycur = self.__ycurAnnotation
        zcur = self.__zcurAnnotation

        if xcur is not None: xcur.enabled = False
        if ycur is not None: ycur.enabled = False
        if zcur is not None: zcur.enabled = False


    def __drawCursorAnnotation(self, canvas, voxel, blockSize=None):
        """Draws the cursor annotation. Highlights the specified voxel with a
        :class:`~fsleyes.gl.annotations.Rect` annotation.

        This is used by mouse motion event handlers, so the user can
        see the possible selection, and thus what would happen if they
        were to click.

        :arg canvas:    The :class:`.SliceCanvas` on which to make the
                        annotation.
        :arg voxel:     Voxel which is at the centre of the cursor.
        :arg blockSize: Size of the cursor square/cube.
        """

        overlay  = self.__currentOverlay
        dopts    = self.displayCtx.getOpts(overlay)
        canvases = [self.__xcanvas,
                    self.__ycanvas,
                    self.__zcanvas]
        cursors  = [self.__xcurAnnotation,
                    self.__ycurAnnotation,
                    self.__zcurAnnotation]

        # If the selected overlay is changed, this
        # method might get called during the overlay
        # changeover (in __selectedOverlayChanged,
        # when the display space warning dialog gets
        # shown). At this point in time, the cursor
        # annotations for the new overlay will not
        # yet have been created, so we can't draw
        # them.
        if any([c is None for c in cursors]):
            return

        # If a block size was not specified,
        # it defaults to selectionSize
        if blockSize is None:
            blockSize = self.selectionSize

        # We need to specify the block
        # size in scaled voxels along
        # each voxel dimension. So we
        # scale the block size by the
        # shortest voxel axis - we're
        # aiming for a square (if 2D)
        # or a cube (if 3D) selection.
        blockSize = np.min(overlay.pixdim[:3]) * blockSize
        blockSize = [blockSize] * 3

        # Limit to the current plane
        # if in 2D selection mode
        if self.selectionIs3D: axes = (0, 1, 2)
        else:                  axes = (canvas.opts.xax, canvas.opts.yax)

        # Calculate a box in the voxel coordinate
        # system, centred at the current voxel,
        # and of the specified block size
        corners = glroutines.voxelBox(voxel,
                                      overlay.shape[:3],
                                      overlay.pixdim[:3],
                                      blockSize,
                                      axes=axes,
                                      bias='high')

        # invalid box size for some reason -
        # maybe the image has bad pixdims?
        if corners is None:
            self.__hideCursorAnnotation()
            return

        # We want the selection to follow voxel
        # edges, but the transformCoords method
        # will map voxel coordinates to the
        # displayed voxel centre. So we offset
        # by -0.5 to get the corners.
        corners = dopts.transformCoords(corners - 0.5, 'voxel', 'display')

        cmin = corners.min(axis=0)
        cmax = corners.max(axis=0)

        for cur, can in zip(cursors, canvases):
            copts = can.opts
            xax   = copts.xax
            yax   = copts.yax
            zax   = copts.zax

            if copts.pos[zax] < cmin[zax] or copts.pos[zax] > cmax[zax]:
                cur.w = 0
                cur.h = 0
                continue

            cur.xy = cmin[[xax, yax]]
            cur.w  = cmax[xax] - cmin[xax]
            cur.h  = cmax[yax] - cmin[yax]

        # Only draw the cursor on the current
        # canvas if locFolMouse is false
        for cur, can in zip(cursors, canvases):
            cur.resetExpiry()
            cur.enabled = can is canvas or self.locationFollowsMouse


    def __refreshCanvases(self):
        """Short cut to refresh the canvases of the :class:`.OrthoPanel`.

        .. note:: This is done instead of calling ``OrthoPanel.Refresh``
                  because the latter introduces flickering.
        """
        self.__xcanvas.Refresh()
        self.__ycanvas.Refresh()
        self.__zcanvas.Refresh()


    def __dynamicRefreshCanvases(self,
                                 ev,
                                 canvas,
                                 mousePos=None,
                                 canvasPos=None):
        """Called by mouse event handlers when the user is interacting with
        a canvas.

        If :attr:`locationFollowsMouse` is ``True``, the
        :attr:`.DisplayContext.location` is set to the current mouse location.
        """

        # If the given location is already
        # equal to the display location,
        # calling _navModeLeftMouseDrag we
        # will not trigger a refresh, so
        # we will force the refresh instead.
        forceRefresh = \
            canvasPos is not None and \
            np.all(np.isclose(canvasPos, self.displayCtx.location.xyz))

        # If locationFollowsMouse==True, we make
        # the canvas location track the edit cursor
        # location, so that the other two canvases
        # update to display the current cursor location.
        if self.locationFollowsMouse and \
           (mousePos  is not None)   and \
           (canvasPos is not None):

            self._navModeLeftMouseDrag(ev, canvas, mousePos, canvasPos)

            if forceRefresh:
                for c in self.getEventTargets():
                    c.Refresh()

        # Otherwise just update the canvas
        # that triggered the event
        else:
            canvas.Refresh()


    def __applySelection(self,
                         canvas,
                         voxel,
                         add=True,
                         combine=False,
                         from_=None):
        """Called by ``sel`` mode mouse handlers. Adds/removes a block
        of voxels, centred at the specified voxel, to/from the current
        :class:`.Selection`.

        :arg canvas:  The source :class:`.SliceCanvas`.

        :arg voxel:   Coordinates of centre voxel.

        :arg add:     If ``True`` a block is added to the selection,
                      otherwise it is removed.

        :arg combine: Tell the :class:`.Selection` object to combine this
                      change with the most recent one.

        :arg from_:   If provided, use the :meth:`.Selection.selectLine`
                      or :meth:`.Selection.deselectLine` methods - should
                      be another voxel coordinate.
        """

        opts = canvas.opts

        if self.selectionIs3D: axes = (0, 1, 2)
        else:                  axes = (opts.xax, opts.yax)

        overlay   = self.__currentOverlay
        editor    = self.__editors[overlay]
        selection = editor.getSelection()
        blockSize = self.selectionSize * np.min(overlay.pixdim[:3])

        if from_ is not None:

            args = (from_, voxel, blockSize, axes, 'high', combine)

            if add: block, offset = selection.selectLine(  *args)
            else:   block, offset = selection.deselectLine(*args)

        else:

            block, offset = glroutines.voxelBlock(
                voxel,
                overlay.shape[:3],
                overlay.pixdim[:3],
                blockSize,
                axes=axes,
                bias='high')

            if add: selection.addToSelection(     block, offset, combine)
            else:   selection.removeFromSelection(block, offset, combine)

        if add: self.__recordSelectionMerger('sel',   offset, block.shape)
        else:   self.__recordSelectionMerger('desel', offset, block.shape)


    def __recordSelectionMerger(self, mode, offset, size):
        """This method is called whenever a change is made to the
        :class:`.Selection` object. It stores some information which is used
        to improve subsequent selection performance when in ``selint`` mode,
        and when :attr:`limitToRadius` is ``True``.


        Basically, if the current selection is limited by radius, and a new,
        similarly limited selection is made, we do not need to clear the
        entire selection before making the new selection - we just need to
        clear the cuboid region in which the previous selection was located.


        This behaviour is referred to as a 'merge' because, ultimately, the
        region of the first selection is merged with the region of the second
        selection, and only this part of the ``Selection`` image is refreshed.


        This method (and the :meth:`__getSelectionMerger` method) contains some
        simple, but awkward, logic which figures out when a merge can happen
        and, conversely, when the full selection does need to be cleared.


        :arg offset: Offset into the selection array of the change.
        :arg size:   Shape of the change.
        """

        # If the user has manually selected anything,
        # we can't merge
        if self.__mergeMode == 'sel':
            return

        self.__mergeMode   = mode
        self.__merge3D     = self.selectionIs3D
        self.__mergeRadius = self.limitToRadius

        # We only care about merging
        # selint+radius blocks
        if mode == 'selint' and self.__mergeRadius:
            self.__mergeBlock  = offset, size


    def __getSelectionMerger(self):
        """This method is called just before a select-by-intensity selection
        is about to happen. It rteturns one of three values:

          - The string ``'clear'``, indicating that the whole selection (or
            the whole slice, if :attr:`selectionIs3D` is ``False``) needs to
            be cleared.

          - The value ``None`` indicating that the selection does not need to
            be cleared, and a merge does not need to be made.

          - A tuple containing the ``(offset, size)`` of a previous change
            to the selection, specifying the portion of the selection which
            needs to be cleared, and which can be subsequently merged with
            a new selection.
        """

        try:
            # If not limiting by radius, the new
            # selectByValue call will clobber the
            # old selection, so we don't need to
            # merge or clear it.
            if not self.limitToRadius:
                return None

            # If the user was selecting voxels,
            # we don't know where those selected
            # voxels are, so we have to clear
            # the full selection.
            if self.__mergeMode == 'sel':
                return 'clear'

            # If the user was just deselecting,
            # we can merge the old block
            if self.__mergeMode == 'desel':
                return self.__mergeBlock

            # If the user was in 2D, but is now
            # in 3D, we have to clear the whole
            # selection. Similarly, if the user
            # was not limiting by radius, but
            # now is, we have to clear.
            if (not self.__merge3D)     and self.selectionIs3D: return 'clear'
            if (not self.__mergeRadius) and self.limitToRadius: return 'clear'

            # Otherwise we can merge the old
            # selection with the new selection.
            return self.__mergeBlock

        finally:
            self.__mergeMode   = None
            self.__merge3D     = None
            self.__mergeRadius = None
            self.__mergeBlock  = None


    def _selModeMouseMove(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse motion events in ``sel`` mode.

        Draws a cursor annotation at the current mouse location
        (see :meth:`__draweCursorAnnotation`).
        """
        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:
            self.__drawCursorAnnotation(canvas, voxel)
            self.__dynamicRefreshCanvases(ev,  canvas)

        return voxel is not None


    def _selModeLeftMouseDown(self,
                              ev,
                              canvas,
                              mousePos,
                              canvasPos,
                              add=True,
                              mode='sel'):
        """Handles mouse down events in ``sel`` mode.

        Starts an :class:`.Editor` change group, and adds to the current
        :class:`Selection`.

        This method is also used by :meth:`_deselModeLeftMouseDown`, which
        may set the ``add`` parameter to ``False``.

        :arg add:  If ``True`` (default) a block at the cursor is added to the
                   selection. Otherwise it is removed.

        :arg mode: The current profile mode (defaults to ``'sel'``).
        """
        if self.__currentOverlay is None:
            return False

        # We clear the Selection object's most
        # recent saved change - all additions
        # to the selection during this click+drag
        # event are merged together (by using
        # the combine flag to addToSelection -
        # see __applySelection). Then, if we are
        # in immediate draw mode, on the
        # up event, we know what part of the
        # selection needs to be refreshed.
        selection = self.__editors[self.__currentOverlay].getSelection()
        selection.setChange(None, None)

        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:
            self.__applySelection(      canvas, voxel, add=add, combine=True)
            self.__drawCursorAnnotation(canvas, voxel)
            self.__dynamicRefreshCanvases(ev,  canvas, mousePos, canvasPos)

        return voxel is not None


    def _selModeLeftMouseDrag(self,
                              ev,
                              canvas,
                              mousePos,
                              canvasPos,
                              add=True,
                              mode='sel'):
        """Handles mouse drag events in ``sel`` mode.

        Adds to the current :class:`Selection`.

        This method is also used by :meth:`_deselModeLeftMouseDown`, which
        may set the ``add`` parameter to ``False``.

        :arg add:  If ``True`` (default) a block at the cursor is added to the
                   selection. Otherwise it is removed.

        :arg mode: The current profile mode (defaults to ``'sel'``).
        """
        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:

            lastPos = self.getLastMouseLocation()[1]
            lastPos = self.__getVoxelLocation(lastPos)

            self.__applySelection(      canvas,
                                        voxel,
                                        add=add,
                                        combine=True,
                                        from_=lastPos)
            self.__drawCursorAnnotation(canvas, voxel)
            self.__dynamicRefreshCanvases(ev,  canvas, mousePos, canvasPos)

        return voxel is not None


    def _selModeLeftMouseUp(
            self, ev, canvas, mousePos, canvasPos, fillValue=None):
        """Handles mouse up events in ``sel`` mode.

        Ends the :class:`.Editor` change group that was started in the
        :meth:`_selModeLeftMouseDown` method.

        This method is also used by :meth:`_deselModeLeftMouseUp`, which
        sets ``fillValue`` to :attr:`eraseValue`.

        :arg fillValue: If :attr:`drawMode` is ``True``, the value to
                        fill the selection with. If not provided, defaults
                        to :attr:`fillValue`.
        """

        if self.__currentOverlay is None:
            return False

        editor    = self.__editors[self.__currentOverlay]
        selection = editor.getSelection()

        # Immediate draw mode - fill
        # and clear the selection.
        if self.drawMode:

            if fillValue is None:
                fillValue = self.fillValue

            editor.fillSelection(fillValue)

            # The Selection object contains the
            # full extent of the changes that
            # were made to the selection during
            # this click+drag event. We only need
            # to clear this part of the selection.
            old, new, off = selection.getLastChange()
            restrict      = [slice(o, o + s) for o, s in zip(off, new.shape)]

            selection.clearSelection(restrict=restrict)

        self.__refreshCanvases()

        return True


    def _selModeMouseLeave(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse leave events in ``sel`` mode. Makes sure that the
        selection cursor annotation is not shown on any canvas.
        """

        self.__hideCursorAnnotation()
        self.__dynamicRefreshCanvases(ev, canvas)


    def _deselModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse down events in ``desel`` mode.

        Calls :meth:`_selModeLeftMouseDown`.
        """
        self._selModeLeftMouseDown(ev,
                                   canvas,
                                   mousePos,
                                   canvasPos,
                                   add=self.drawMode,
                                   mode='desel')


    def _deselModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse drag events in ``desel`` mode.

        Calls :meth:`_selModeLeftMouseDrag`.
        """
        self._selModeLeftMouseDrag(ev,
                                   canvas,
                                   mousePos,
                                   canvasPos,
                                   add=self.drawMode,
                                   mode='desel')


    def _deselModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse up events in ``desel`` mode.

        Calls :meth:`_selModeLeftMouseUp`.
        """
        self._selModeLeftMouseUp(
            ev, canvas, mousePos, canvasPos, fillValue=self.eraseValue)


    def _chsizeModeMouseWheel(self, ev, canvas, wheelDir, mousePos, canvasPos):
        """Handles mouse wheel events in ``chsize`` mode.

        Increases/decreases the current :attr:`selectionSize`.
        """

        if   wheelDir > 0: self.selectionSize += 1
        elif wheelDir < 0: self.selectionSize -= 1

        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is None:
            return False

        # See comment in OrthoViewProfile._zoomModeMouseWheel
        # about timeout
        def update():
            self.__drawCursorAnnotation(canvas, voxel)
            self.__dynamicRefreshCanvases(ev, canvas)

        idle.idle(update, timeout=0.1)

        return True


    def __selintPropertyChanged(self, *a):
        """Called when the :attr:`intensityThres`, :attr:`localFill`,
        :attr:`limitToRadius`, or :attr:`searchRadius` properties change.
        Re-runs select-by-intensity (via :meth:`__selintSelect`), with
        the new settings.
        """

        if self.__currentOverlay is None:
            return

        # If the last mouse up was not
        # select-by-intensity, don't
        # adjust the selection
        if self.getLastMouseUpHandler()[0] != 'selint':
            return

        mousePos, canvasPos = self.getLastMouseUpLocation()
        canvas              = self.getLastCanvas()

        if mousePos is None or canvas is None:
            return

        voxel = self.__getVoxelLocation(canvasPos)

        def update():
            self.__selintSelect(voxel, canvas)
            self.__refreshCanvases()

        if voxel is not None:

            # Asynchronously update the select-by-intensity
            # selection - we do it async, and with a time out,
            # so we don't queue loads of redundant jobs while
            # the user is e.g. dragging the intensityThres
            # slider real fast.
            idle.idle(update, timeout=0.1)


    def __selintThresLimitChanged(self, *a):
        """Called when the :attr:`intensityThresLimit` changes. Updates the
        maximum value on the :attr:`intensityThres` accordingly.
        """
        self.setAttribute('intensityThres',
                          'maxval',
                          self.intensityThresLimit)


    def __selintSelect(self, voxel, canvas):
        """Selects voxels by intensity, using the specified ``voxel`` as
        the seed location.

        Called by the :meth:`_selintModeLeftMouseDown`,
        :meth:`_selintModeLeftMouseDrag`,
        :meth:`_selintModeLeftMouseWheel`, and :meth:`__selintPropertyChanged`
        methods.  See :meth:`.Selection.selectByValue`.
        """

        overlay = self.__currentOverlay

        if overlay is None:
            return False

        editor = self.__editors[overlay]

        # The searchRadius/intensityThres
        # properties are unclamped, so we
        # have to make sure that they have
        # valid values.
        if self.searchRadius   > 0: searchRadius   = self.searchRadius
        else:                       searchRadius   = 0
        if self.intensityThres > 0: intensityThres = self.intensityThres
        else:                       intensityThres = 0

        if not self.limitToRadius:
            searchRadius = None
        else:
            searchRadius = (searchRadius / overlay.pixdim[0],
                            searchRadius / overlay.pixdim[1],
                            searchRadius / overlay.pixdim[2])

        if self.selectionIs3D:
            restrict = None
        else:
            zax           = canvas.opts.zax
            restrict      = [slice(None, None, None) for i in range(3)]
            restrict[zax] = slice(voxel[zax], voxel[zax] + 1)

        # We may need to manually clear part or all
        # of the selection before running the select
        # by value routine. The get/recordSelectionMerger
        # methods take care of the logic needed to
        # figure out what we need to do.
        selection = editor.getSelection()
        merge     = self.__getSelectionMerger()

        # The whole selection/slice needs clearing.
        # We suppress any notification by the Selection
        # object at this point - notification will
        # happen via the selectByValue method call below.
        if merge == 'clear':
            with selection.skipAll():
                selection.clearSelection(restrict=restrict)

        # We only need to clear a region
        # within the selection
        elif merge is not None:

            # Note that we are telling the
            # selectByValuem method below
            # 'combine' any previous selection
            # change with the new one, This
            # means that the entire selection
            # image is going to be replaced
            with selection.skipAll():

                # If we're in 2D mode, we just clear
                # the whole slice, as it should be fast
                # enough.
                if not self.selectionIs3D:

                    selection.clearSelection(restrict=restrict)

                # Otherwise we just clear the region
                else:
                    off, size  = merge
                    clearBlock = [slice(o, o + s) for o, s in zip(off, size)]

                    selection.clearSelection(restrict=clearBlock)

        # The 'combine' flag tells the selection object
        # to merge the last change (the clearSelection
        # call above) with the new change, so that the
        # Selection.getLastChange method will return
        # the union of those two regions.
        #
        # This is important, because the SelectionTexture
        # object, which is listening to changes on the
        # Selection object, will only need to update that
        # part of the GL texture.
        selected, offset = selection.selectByValue(
            voxel,
            precision=intensityThres,
            searchRadius=searchRadius,
            local=self.localFill,
            restrict=restrict,
            combine=merge is not None)

        self.__recordSelectionMerger('selint', offset, selected.shape)

        return True


    def _selintModeMouseMove(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse motion events in ``selint`` mode. Draws a selection
        annotation at the current location (see
        :meth:`__drawCursorAnnotation`).
        """
        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:
            self.__drawCursorAnnotation(canvas, voxel, 1)
            self.__dynamicRefreshCanvases(ev,  canvas)

        return voxel is not None


    def _selintModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse down events in ``selint`` mode.

        Starts an :class:`.Editor` change group, then clears the current
        selection, and selects voxels by intensity (see
        :meth:`__selintSelect`).
        """

        if self.__currentOverlay is None:
            return False

        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:
            self.__selintSelect(voxel, canvas)
            self.__dynamicRefreshCanvases(ev, canvas, mousePos, canvasPos)

        return voxel is not None


    def _selintModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse drag events in ``selint`` mode.

        A select-by-intensity is re-run with the current mouse location.  See
        the :meth:`__selintSelect` method.
        """

        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:

            refreshArgs = (ev, canvas, mousePos, canvasPos)

            self.__drawCursorAnnotation(canvas, voxel, 1)
            self.__selintSelect(voxel, canvas)
            self.__dynamicRefreshCanvases(*refreshArgs)

        return voxel is not None


    def _selintModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse up events in ``selint`` mode. Ends the :class:`.Editor`
        change group that was started in the :meth:`_selintModeLeftMouseDown`
        method.
        """
        if self.__currentOverlay is None:
            return False

        self.__refreshCanvases()

        return True


    def _selintModeMouseLeave(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse leave events in ``selint`` mode. Makes sure that
        the selection cursor annotation is not shown on any canvas.
        """
        self.__hideCursorAnnotation()
        self.__dynamicRefreshCanvases(ev, canvas)


    def _fillModeMouseMove(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse motion events in ``fill`` mode. Draws a selection
        annotation at the current location (see
        :meth:`__drawCursorAnnotation`).
        """
        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is not None:
            self.__drawCursorAnnotation(canvas, voxel, 1)
            self.__dynamicRefreshCanvases(ev,  canvas)

        return voxel is not None


    def _fillModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Handles mouse down events in ``fill`` mode. Calls
        :meth:`.Selection.invertRegion` at the current location.
        """

        if self.__currentOverlay is None:
            return False

        voxel = self.__getVoxelLocation(canvasPos)

        if voxel is None:
            return False

        zax           = canvas.opts.zax
        restrict      = [slice(None, None, None) for i in range(3)]
        restrict[zax] = slice(voxel[zax], voxel[zax] + 1)
        editor        = self.__editors[self.__currentOverlay]
        selection     = editor.getSelection()

        # draw mode - works in essentially the
        # same manner as select-by-intensity,
        # but we use a threshold of 0.5
        # (intended for binary masks), and
        # immediately fill the selected region.
        if self.drawMode:
            editor.startChangeGroup()
            selected, offset = selection.selectByValue(
                voxel,
                precision=0.5,
                local=True,
                restrict=restrict)
            editor.fillSelection(self.fillValue)
            selection.clearSelection(restrict=restrict)
            editor.endChangeGroup()

        # select mode - we invert the select
        # state of the clicked-on region.
        else:
            selection.invertRegion(voxel, restrict=restrict)

        self.__dynamicRefreshCanvases(ev, canvas, mousePos, canvasPos)

        return True


    def _chthresModeMouseWheel(self, ev, canvas, wheel, mousePos, canvasPos):
        """Handles mouse wheel events in ``chthres`` mode.

        The :attr:`intensityThres` value is decreased/increased according to
        the mouse wheel direction. If the mouse button is down,
        select-by-intensity is re-run at the current mouse location.
        """
        overlay   = self.displayCtx.getSelectedOverlay()
        dataRange = overlay.dataRange[1] - overlay.dataRange[0]
        step      = 0.01 * dataRange

        if   wheel > 0: offset =  step
        elif wheel < 0: offset = -step
        else:           return False

        self.intensityThres += offset

        return True


    def _chradModeMouseWheel(self, ev, canvas, wheel, mousePos, canvasPos):
        """Handles mouse wheel events in ``chrad`` mode.

        The :attr:`searchRadius` value is decreased/increased according
        to the mouse wheel direction. If the mouse button is down,
        select-by-intensity is re-run at the current mouse location.
        """

        if   wheel > 0: offset =  2.5
        elif wheel < 0: offset = -2.5
        else:           return False

        self.searchRadius += offset

        return True



#!/usr/bin/env python
#
# atlasoverlaypanel.py -  The AtlasOverlayPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AtlasOverlayPanel`, which is a sub-panel
used by the :class:`.AtlasPanel`.
"""


import logging

import wx

import fsl.data.atlases                     as atlases
import fsl.utils.idle                       as idle
from   fsl.utils.platform import platform   as fslplatform

import fsleyes_widgets.elistbox             as elistbox
import fsleyes_widgets.placeholder_textctrl as plctext
import fsleyes_widgets.utils.status         as status
import fsleyes.panel                        as fslpanel
import fsleyes.strings                      as strings



log = logging.getLogger(__name__)


class AtlasOverlayPanel(fslpanel.FSLeyesPanel):
    """The ``AtlasOverlayPanel`` displays a list of all available FSL atlases
    (see the :mod:`.atlases` module), and allows the user to:

      1. Search across all atlases for regions, by name.

      2. Toggle overlays for atlases, and for individual regions from any
         atlas.

      3. Move the :attr:`.DisplayContext.location` to any region from any
         atlas.


    An ``AtlasOverlayPanel`` looks something like this:

    .. image:: images/atlasoverlaypanel.png
       :scale: 50%
       :align: center


    The ``AtlasOverlayPanel`` has three main sections:

      - The *atlas list* - a :class:`fsleyes_widgets.elistbox.EditableListBox`
        with an entry, and an :class:`OverlayListWidget` for every available
        atlas.  The ``OverlayListWidget`` allows the user to turn on/off a
        summary overlay image for the atlas (see the section on :ref:`atlas
        panel overlays <atlas-panel-atlas-overlays>`).

      - The *region list* - an ``EditableListBox`` which contains one entry
        for every region in the atlas that is currently selected in the atlas
        list, and an ``OverlayListWidget`` alongside every region. The
        ``OverlayListWidget`` allows the user to navigate to a region, and
        to turn on/off a label or statistic/probabilistic region overlay.

      - The *region filter* - a ``wx.TextCtrl`` located above the region
        list. This allows the user to search for regions by name. When the
        user types a string into this field, the region list will be updated
        to show only regions that match the entered string (via a simple
        substring match), and the atlas list will be updated such that all
        atlases which have matching regions will be highlighted in bold.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, atlasPanel):
        """Create an ``AtlasOverlayPanel``.

        :arg parent:      the :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` instance.

        :arg atlasPanel:  The :class:`.AtlasPanel` instance that has created
                          this ``AtlasInfoPanel``.
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        # References to an EditableListBox
        # for each atlas, containing a list
        # of its regions. These are created
        # on-demand in the __onAtlasSelect
        # method.
        self.__regionLists = {}

        self.__atlasPanel      = atlasPanel
        self.__contentPanel    = wx.SplitterWindow(self,
                                                   style=wx.SP_LIVE_UPDATE)
        self.__atlasList       = elistbox.EditableListBox(
            self.__contentPanel,
            vgap=5,
            style=(elistbox.ELB_NO_ADD    |
                   elistbox.ELB_NO_REMOVE |
                   elistbox.ELB_NO_MOVE))

        self.__regionPanel     = wx.Panel(self.__contentPanel)
        self.__regionFilter    = plctext.PlaceholderTextCtrl(
            self.__regionPanel, placeholder='Search')

        self.__contentPanel.SetMinimumPaneSize(50)
        self.__contentPanel.SplitVertically(self.__atlasList,
                                            self.__regionPanel)
        self.__contentPanel.SetSashGravity(0.4)

        self.__sizer       = wx.BoxSizer(wx.HORIZONTAL)
        self.__regionSizer = wx.BoxSizer(wx.VERTICAL)

        self.__regionSizer.Add(self.__regionFilter, flag=wx.EXPAND)
        self.__regionSizer.AddStretchSpacer()

        self.__sizer      .Add(self.__contentPanel,
                               flag=wx.EXPAND,
                               proportion=1)

        self.__regionPanel.SetSizer(self.__regionSizer)
        self              .SetSizer(self.__sizer)

        self.__regionFilter.Bind(wx.EVT_TEXT, self.__onRegionFilter)
        self.__atlasList.Bind(elistbox.EVT_ELB_SELECT_EVENT,
                              self.__onAtlasSelect)

        fslplatform.register(     self.name, self.__fslDirChanged)
        atlases.registry.register(self.name, self.__atlasAdded,   'add')
        atlases.registry.register(self.name, self.__atlasRemoved, 'remove')

        self.__buildAtlasList()

        self.__regionSizer.Layout()
        self.__sizer      .Layout()

        # Allow the atlas list
        # to be minimised
        self.__atlasList.SetMinSize((50, -1))

        self.SetMinSize(self.__sizer.GetMinSize())


    def destroy(self):
        """Performs some clean up operations. """
        fslplatform.deregister(self.name)
        fslpanel.FSLeyesPanel.destroy(self)


    def Enable(self, enable=True):
        """Enables/disables this ``AtlasOverlayPanel``. """

        self.__atlasList.Enable(enable)

        for atlasID, regionList in self.__regionLists.items():
            regionList.Enable(enable)


    def Disable(self):
        """Disables this ``AtlasOverlayPanel``. """
        self.Enable(False)


    def setOverlayState(self, atlasDesc, labelIdx, summary, state):
        """Updates the *enabled* state of the specified atlas overlay.

        This method is called by the :class:`.AtlasPanel` when an atlas overlay
        is added/removed to/from the :class:`.OverlayList`. It ensures that the
        :class:`OverlayListWidget` controls for every atlas, and every region,
        are up to date.
        """

        atlasIdx = self.__atlasList.IndexOf(atlasDesc)

        log.debug('Setting {}/{} overlay state to {}'.format(
            atlasDesc.atlasID, labelIdx, state))

        if labelIdx is None:
            widget = self.__atlasList.GetItemWidget(atlasIdx)
            widget.SetEnableState(state)
        else:

            regionList = self.__regionLists.get(atlasDesc.atlasID, None)
            if regionList is not None:
                regionList.GetItemWidget(labelIdx).SetEnableState(state)


    def __fslDirChanged(self, *a):
        """Called when the :attr:`.Platform.fsldir` changes. Refreshes
        the atlas list.
        """
        self.__buildAtlasList()


    def __atlasAdded(self, *a):
        """Called when a new atlas is added to the :class:`.AtlasRegistry`.
        Re-generates the atlas list.
        """
        self.__buildAtlasList()


    def __atlasRemoved(self, *a):
        """Called when an atlas is removed from the :class:`.AtlasRegistry`.
        Re-generates the atlas list.
        """
        self.__buildAtlasList()


    def __buildAtlasList(self):
        """Clears and recreates the atlas list. Also clears all existing
        region lists.
        """

        atlasDescs = atlases.listAtlases()

        # This method is called whenever any
        # atlases are added/removed. We want
        # to preserve any region lists for
        # atlases that are still in the atlas
        # registry.
        regionLists = dict(self.__regionLists)

        # If a region list is currently
        # being shown, clear it.
        regionList = self.__regionSizer.GetItem(1).GetWindow()
        if regionList is not None:
            regionList.Show(False)
            self.__regionSizer.Remove(1)
            self.__regionSizer.AddStretchSpacer()

        self.__regionLists = {}

        # Now clear and re-populate the atlas list
        self.__atlasList.Clear()
        for i, atlasDesc in enumerate(atlasDescs):

            self.__atlasList.Append(atlasDesc.name, atlasDesc)
            self.__updateAtlasState(atlasDesc)

            widget = OverlayListWidget(self.__atlasList,
                                       atlasDesc.atlasID,
                                       self.__atlasPanel,
                                       self)

            self.__atlasList.SetItemWidget(i, widget)

        # Restore references to region lists
        # for atlases that still exist
        for atlasID, regionList in regionLists.items():
            if atlases.hasAtlas(atlasID):
                self.__regionLists[atlasID] = regionList
            else:
                regionList.Destroy()

        self.__regionSizer.Layout()


    def __onRegionFilter(self, ev):
        """Called when the user enters some text in the region filter.

        Filters the region list (see the :meth:`.EditableListBox.ApplyFilter`
        method), and updates all items in the atlas list (see the
        :meth:`__updateAtlasState` method).
        """

        filterStr = self.__regionFilter.GetValue().lower().strip()

        for atlasDesc in atlases.listAtlases():

            self.__updateAtlasState(atlasDesc)

            listBox = self.__regionLists.get(atlasDesc.atlasID, None)

            if listBox is not None:
                listBox.ApplyFilter(filterStr, ignoreCase=True)


    def __updateAtlasState(self, atlasDesc):
        """Updates the state of the atlas list item which corresponds to the
        atlas with the specified identifier.

        If the atlas has regions which match the current region filter string,
        the atlas label font is set to bold. Otherwise (or if the region filter
        string is empty), the atlas label font is set to normal.
        """

        atlasIdx  = self.__atlasList.IndexOf(atlasDesc)
        filterStr = self.__regionFilter.GetValue().lower().strip()
        atlasDesc = self.__atlasList.GetItemData(atlasIdx)

        if filterStr == '':
            nhits = 0
        else:
            nhits = len([l for l in atlasDesc.labels
                         if filterStr in l.name.lower()])

        if nhits == 0:
            weight = wx.FONTWEIGHT_LIGHT
            colour = '#404040'
        else:
            weight = wx.FONTWEIGHT_BOLD
            colour = '#000000'

        font = self.__atlasList.GetItemFont(atlasIdx)
        font.SetWeight(weight)

        self.__atlasList.SetItemFont(atlasIdx, font)
        self.__atlasList.SetItemForegroundColour(atlasIdx, colour, colour)


    def __onAtlasSelect(self, ev=None, atlasDesc=None):
        """Called when the user selects an atlas in the atlas list, or
        the :meth:`selectAtlas` method is called.

        If a region list (a list of :class:`OverlayListWidget` items for every
        region in the atlas, to be displayed in the region list) has not yet
        been created, it is created - this is done asynchronously (via the
        :func:`idle.idle` function), as it can take quite a long time for
        some of the atlases (e.g. the Talairach and Juelich).

        Then the region list is updated to show the regions for the newly
        selected atlas.
        """

        if ev is not None:
            atlasDesc = ev.data

        regionList         = self.__regionLists.get(atlasDesc.atlasID, None)
        atlasPanelDisabled = regionList is None

        # This function changes the displayed region
        # list. We schedule it on the wx idle loop,
        # so it will get called after the region list
        # has been populated (if it has not been
        # displayed before).
        def changeAtlasList():

            # See comment above about
            # suppressing wx complaints
            try:

                filterStr = self.__regionFilter.GetValue().lower().strip()
                regionList.ApplyFilter(filterStr, ignoreCase=True)

                self.__updateAtlasState(atlasDesc)

                status.update(strings.messages[self, 'regionsLoaded'].format(
                    atlasDesc.name))
                log.debug('Showing region list for {} ({})'.format(
                    atlasDesc.atlasID, id(regionList)))

                # Hide the currently
                # shown region list
                old = self.__regionSizer.GetItem(1).GetWindow()
                if old is not None:
                    old.Show(False)

                regionList.Show(True)
                self.__regionSizer.Remove(1)
                self.__regionSizer.Insert(1,
                                          regionList,
                                          flag=wx.EXPAND,
                                          proportion=1)
                self.__regionSizer.Layout()

                if atlasPanelDisabled:
                    self.__atlasPanel.enableAtlasPanel()

            except RuntimeError:
                pass

        if regionList is None:

            # The region list for this atlas has not yet been
            # created. So we create the list, and then create
            # a widget for every region in the atlas. Some of
            # the atlases (Juelich and Talairach in particular)
            # have a large number of regions, so we create the
            # widgets asynchronously on the wx idle loop.
            regionList = elistbox.EditableListBox(
                self.__regionPanel,
                vgap=5,
                style=(elistbox.ELB_NO_ADD    |
                       elistbox.ELB_NO_REMOVE |
                       elistbox.ELB_NO_MOVE))
            regionList.Show(False)

            self.__regionLists[atlasDesc.atlasID] = regionList

            # Add blockSize labels, starting from label[i],
            # to the region list. Then, if necessary,
            # schedule more labels be added, starting from
            # label[i + blockSize].
            blockSize = 20
            nlabels   = len(atlasDesc.labels)

            def addToRegionList(start):

                # If the user kills this panel while
                # the region list is being updated,
                # suppress wx complaints.
                try:

                    for i in range(start, min(start + blockSize, nlabels)):
                        label = atlasDesc.labels[i]
                        widget = OverlayListWidget(regionList,
                                                   atlasDesc.atlasID,
                                                   self.__atlasPanel,
                                                   self,
                                                   label.index)
                        regionList.Append(label.name, extraWidget=widget)

                    if i < nlabels - 1: idle.idle(addToRegionList, i + 1)
                    else:               idle.idle(changeAtlasList)

                except RuntimeError:
                    pass

            log.debug('Creating region list for {} ({})'.format(
                atlasDesc.atlasID, id(regionList)))

            status.update(
                strings.messages[self, 'loadRegions'].format(atlasDesc.name),
                timeout=None)

            # Schedule addToRegionList on the
            # wx idle loop for the first region.
            # The function will recursively
            # schedule itself to run for subsequent
            # regions.
            #
            # Disable the panel while this is
            # occurring.

            atlasPanelDisabled = True

            self.__atlasPanel.enableAtlasPanel(False)
            idle.idle(addToRegionList, 0)
        else:
            idle.idle(changeAtlasList)


    def selectAtlas(self, atlasDesc):
        """Selects the specified atlas. This method is used by
        :class:`OverlayListWidget` instances.

        :arg atlasID:   The atlas identifier
        """

        atlasIdx = self.__atlasList.IndexOf(atlasDesc)

        self.__atlasList.SetSelection(atlasIdx)
        self.__onAtlasSelect(atlasDesc=atlasDesc)


class OverlayListWidget(wx.Panel):
    """``OverlayListWidget`` items are used by the :class:`AtlasOverlayPanel`
    in both the atlas list and the region list.

    For atlases, an ``OverlayListWidget`` contains a check box allowing the
    user to toggle the visibility of a summary (i.e. label) overlay for that
    atlas.

    For regions, an ``OverlayListWidget`` contains a check box and a button.
    The check box allows the user to toggle visibility of an atlas region
    overlay, and the button calls the :meth:`.AtlasPanel.locateRegion`
    method, allowing the user to locate the region.

    See the :class:`.AtlasPanel` class documentation, and the :mod:`.atlases`
    module for more details
    """


    def __init__(self,
                 parent,
                 atlasID,
                 atlasPanel,
                 atlasOvlPanel,
                 labelIdx=None):
        """Create an ``OverlayListWidget``.

        :arg parent:        The :mod:`wx` parent object - this is assumed to
                            be an :class:`.EditableListBox`.

        :arg atlasID:       The atlas identifier.

        :arg atlasOvlPanel: The :class:`AtlasOverlayPanel` which created this
                            ``OverlayListWidget``.

        :arg atlasPanel:    The :class:`.AtlasPanel` which owns the
                            :class:`AtlasOverlayPanel` that created this
                            ``OverlayListWidget``.

        :arg labelIdx:      Label index of the region, if this
                            ``OverlatyListWidget`` corresponds to a region,
                            or ``None``  if it corresponds to an atlas.
        """

        wx.Panel.__init__(self, parent)

        self.__atlasID       = atlasID
        self.__atlasDesc     = atlases.getAtlasDescription(atlasID)
        self.__atlasPanel    = atlasPanel
        self.__atlasOvlPanel = atlasOvlPanel
        self.__atlasList     = parent
        self.__labelIdx      = labelIdx

        self.__enableBox = wx.CheckBox(self)
        self.__enableBox.SetValue(False)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__enableBox, flag=wx.EXPAND)

        self.__enableBox.Bind(wx.EVT_CHECKBOX, self.__onEnable)

        if labelIdx is not None:
            self.__locateButton = wx.Button(self,
                                            label='+',
                                            style=wx.BU_EXACTFIT)
            self.__sizer.Add(self.__locateButton, flag=wx.EXPAND)

            self.__locateButton.Bind(wx.EVT_BUTTON, self.__onLocate)


    def SetEnableState(self, state):
        """Sets the enable overlay checkbox state to the specified state
        (``True`` or ``False``).
        """
        self.__enableBox.SetValue(state)


    def __onEnable(self, ev):
        """Called when the enable overlay checkbox is clicked. Calls the
        :meth:`.AtlasPanel.toggleOverlay` method, to toggle the overlay
        for the atlas/region associated with this ``OverlayListWidget``..
        """

        def onLoad():

            if not self or \
               not self.__atlasOvlPanel or \
               self.__atlasOvlPanel.destroyed:
                return

            self.__atlasPanel.enableAtlasPanel()

        def onError(e):
            message = strings.messages[self.__atlasOvlPanel, 'loadAtlasError']
            message = message.format(
                self.__atlasID, '{} ({})'.format(type(e).__name__, str(e)))
            wx.MessageDialog(
                self.GetTopLevelParent(),
                message=message,
                style=(wx.ICON_EXCLAMATION | wx.OK)).ShowModal()

            self.__atlasPanel.enableAtlasPanel()
            self.__enableBox.SetValue(
                self.__atlasPanel.getOverlayState(
                    self.__atlasID,
                    self.__labelIdx,
                    self.__atlasDesc.atlasType == 'label'))

        self.__atlasPanel.enableAtlasPanel(False)

        log.debug('Toggling atlas {}'.format(self.__atlasID))

        self.__atlasPanel.toggleOverlay(
            self.__atlasID,
            self.__labelIdx,
            self.__atlasDesc.atlasType == 'label',
            onLoad=onLoad,
            onError=onError)

        if self.__labelIdx is None:
            self.__atlasOvlPanel.selectAtlas(self.__atlasDesc)


    def __onLocate(self, ev):
        """Called when the locate region button is clicked only on
        ``OverlayListWidget`` items which are associated with an atlas
        region).

        Calls the :meth:`.AtlasPanel.locateRegion` method for the atlas/region
        associated with this ``OverlayListWidget``.
        """
        self.__atlasPanel.locateRegion(self.__atlasID, self.__labelIdx)



#!/usr/bin/env python
#
# powerspectrumseries.py - Classes used by the PowerSpectrumPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides :class:`.DataSeries` sub-classes which are used
by the :class:`.PowerSpectrumPanel` for plotting power spectra.

The following classes are provided:

.. autosummary::
   :nosignatures:

   PowerSpectrumSeries
   VoxelPowerSpectrumSeries
   ComplexPowerSpectrumSeries
   ImaginaryPowerSpectrumSeries
   MagnitudePowerSpectrumSeries
   PhasePowerSpectrumSeries
   MelodicPowerSpectrumSeries
"""


import logging

import numpy     as np
import numpy.fft as fft

import fsl.data.image        as fslimage
import fsl.data.melodicimage as fslmelimage
import fsleyes_props         as props
import fsleyes.colourmaps    as fslcm
import fsleyes.strings       as strings
from . import                   dataseries


log = logging.getLogger(__name__)


def calcPowerSpectrum(data, varNorm=False):
    """Calculates a power spectrum for the given one-dimensional data array. If
    ``varNorm is True``, the data is de-meaned and normalised by its standard
    deviation before the fourier transformation.

    :arg data:    Numpy array containing the time series data

    :arg varNorm: Normalise the data before fourier transformation

    :returns:     If ``data`` contains real values, the magnitude of the power
                  spectrum is returned. If ``data`` contains complex values,
                  the complex power spectrum is returned.
    """

    # De-mean and normalise
    # by standard deviation
    if varNorm:
        mean = data.mean()
        std  = data.std()

        if not np.isclose(std, 0):
            data = data - mean
            data = data / std

        # If all values in the data
        # are the same, it has mean 0
        else:
            data = np.zeros(data.shape, dtype=data.dtype)

    # Fourier transform on complex data
    if np.issubdtype(data.dtype, np.complexfloating):
        data = fft.fft(data)
        data = fft.fftshift(data)

    # Fourier transform on real data - we
    # calculate and return the magnitude.
    # We also drop the first (zero-frequency)
    # term (see the rfft docs) as it is
    # useless when varNorm is disabled.
    else:
        data = fft.rfft(data)[1:]
        data = magnitude(data)

    return data


def calcFrequencies(data, sampleTime):
    """Calculates the frequencies of the power spectrum for the given
    data.

    :arg data:       The input time series data
    :arg sampleTime: Time between each data point
    :returns:        A ``numpy`` array containing the frequencies of the
                     power spectrum for ``data``
    """

    nsamples = len(data)

    if np.issubdtype(data.dtype, np.complexfloating):
        xdata = fft.fftfreq(nsamples, sampleTime)
        xdata = fft.fftshift(xdata)
    else:
        # Drop the zero-frequency term
        # (see calcPowerSpectrum)
        xdata = fft.rfftfreq(nsamples, sampleTime)[1:]

    return xdata


def magnitude(data):
    """Returns the magnitude of the given complex data. """
    real  = data.real
    imag  = data.imag
    return np.sqrt(real ** 2 + imag ** 2)


def phase(data):
    """Returns the phase of the given complex data. """
    real = data.real
    imag = data.imag
    return np.arctan2(imag, real)


def phaseCorrection(spectrum, freqs, p0, p1):
    """Applies phase correction to the given complex power spectrum.

    :arg spectrum: Complex-valued power spectrum
    :arg freqs:    Spectrum frequency bins
    :arg p0:       Zero order phase correction term
    :arg p1:       First order phase correction term
    :returns:      The corrected power spectrum.
    """
    exp = 1j * 2 * np.pi * (p0 / 360 + freqs * p1)
    return np.exp(exp) * spectrum


class PowerSpectrumSeries(object):
    """The ``PowerSpectrumSeries`` encapsulates a power spectrum data series
    from an overlay. The ``PowerSpectrumSeries`` class is a base mixin class
    for all other classes in this module.
    """


    varNorm  = props.Boolean(default=True)
    """If ``True``, the data is normalised to unit variance before the fourier
    transformation.
    """


    @property
    def sampleTime(self):
        """Returns the time between time series samples for the overlay
        data. """
        if isinstance(self.overlay, fslmelimage.MelodicImage):
            return self.overlay.tr
        elif isinstance(self.overlay, fslimage.Image):
            return self.overlay.pixdim[3]
        else:
            return 1


class VoxelPowerSpectrumSeries(dataseries.VoxelDataSeries,
                               PowerSpectrumSeries):
    """The ``VoxelPowerSpectrumSeries`` class encapsulates the power spectrum
    of a single voxel from a 4D :class:`.Image` overlay. The voxel is dictated
    by the :attr:`.DisplayContext.location` property.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``VoxelPowerSpectrumSeries``.  All arguments are passed
        to the :meth:`VoxelDataSeries.__init__` method. A :exc:`ValueError`
        is raised if the overlay is not a 4D :class:`.Image`.
        """

        dataseries.VoxelDataSeries.__init__(self, *args, **kwargs)

        if self.overlay.ndim < 4:
            raise ValueError('Overlay is not a 4D image')


    def getData(self):
        """Returns the data at the current voxel. """

        data = self.dataAtCurrentVoxel()

        if data is None:
            return None, None

        xdata = calcFrequencies(  data, self.sampleTime)
        ydata = calcPowerSpectrum(data, self.varNorm)

        return xdata, ydata


class ComplexPowerSpectrumSeries(VoxelPowerSpectrumSeries):
    """This class is the frequency-spectrum equivalent of the
    :class:`.ComplexTimeSeries` class - see it for more details.
    """

    plotReal      = props.Boolean(default=True)
    plotImaginary = props.Boolean(default=False)
    plotMagnitude = props.Boolean(default=False)
    plotPhase     = props.Boolean(default=False)


    zeroOrderPhaseCorrection = props.Real(default=0)
    """Apply zero order phase correction to the power spectrum of the complex
    data.
    """


    firstOrderPhaseCorrection = props.Real(default=0)
    """Apply first order phase correction to the power spectrum of the complex
    data.
    """


    def __init__(self, overlay, overlayList, displayCtx, plotPanel):
        """Create a ``ComplexPowerSpectrumSeries``. All arguments are
        passed through to the :class:`VoxelPowerSpectrumSeries` constructor.
        """

        VoxelPowerSpectrumSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)

        self.__cachedData = (None, None)
        self.__imagps     = ImaginaryPowerSpectrumSeries(
            self, overlay, overlayList, displayCtx, plotPanel)
        self.__magps      = MagnitudePowerSpectrumSeries(
            self, overlay, overlayList, displayCtx, plotPanel)
        self.__phaseps    = PhasePowerSpectrumSeries(
            self, overlay, overlayList, displayCtx, plotPanel)

        for ps in (self.__imagps, self.__magps, self.__phaseps):
            ps.colour = fslcm.randomDarkColour()
            ps.bindProps('alpha',     self)
            ps.bindProps('lineWidth', self)
            ps.bindProps('lineStyle', self)


    def makeLabel(self):
        """Returns a string representation of this
        ``ComplexPowerSpectrumSeries`` instance.
        """
        return '{} ({})'.format(VoxelPowerSpectrumSeries.makeLabel(self),
                                strings.labels[self])


    @property
    def cachedData(self):
        """Returns the currently cached data (see :meth:`getData`). """
        return self.__cachedData


    def getData(self):
        """If :attr:`plotReal` is true, returns the real component of the power
        spectrum of the data at the current voxel. Otherwise returns ``(None,
        None)``.

        Every time this method is called, the power spectrum is calculated,
        phase correction is applied, and a reference to the resulting complex
        power spectrum (and frequencies) is saved; it is accessible via the
        :meth:`cachedData` property, for use by the
        :class:`ImaginaryPowerSpectrumSeries`,
        :class:`MagnitudePowerSpectrumSeries`, and
        :class:`PhasePowerSpectrumSeries`.
        """

        xdata, ydata = VoxelPowerSpectrumSeries.getData(self)

        if self.zeroOrderPhaseCorrection  != 0 or \
           self.firstOrderPhaseCorrection != 0:
            ydata = phaseCorrection(ydata,
                                    xdata,
                                    self.zeroOrderPhaseCorrection,
                                    self.firstOrderPhaseCorrection)

        # Note that we're assuming that this
        # ComplexPowerSpectrumSeries.getData
        # method will be called before the
        # corresponding call(s) to the
        # Imaginary/Magnitude/Phase series
        # methods.
        self.__cachedData = xdata, ydata

        if not self.plotReal:
            return None, None

        if ydata is not None:
            ydata = ydata.real

        return xdata, ydata


    def extraSeries(self):
        """Returns a list of additional series to be plotted, based
        on the values of the :attr:`plotImaginary`, :attr:`plotMagnitude`
        and :attr:`plotPhase` properties.
        """

        extras = []
        if self.plotImaginary: extras.append(self.__imagps)
        if self.plotMagnitude: extras.append(self.__magps)
        if self.plotPhase:     extras.append(self.__phaseps)
        return extras


class ImaginaryPowerSpectrumSeries(dataseries.DataSeries):
    """An ``ImaginaryPowerSpectrumSeries`` represents the power spectrum of the
    imaginary component of a complex-valued image.
    ``ImaginaryPowerSpectrumSeries`` instances are created by
    :class:`ComplexPowerSpectrumSeries` instances.
    """


    def __init__(self, parent, *args, **kwargs):
        """Create an ``ImaginaryPowerSpectrumSeries``.

        :arg parent: The :class:`ComplexPowerSpectrumSeries` which owns this
                     ``ImaginaryPowerSpectrumSeries``.

        All other arguments are passed through to the :class:`DataSeries`
        constructor.
        """
        dataseries.DataSeries.__init__(self, *args, **kwargs)
        self.__parent = parent


    def makeLabel(self):
        """Returns a string representation of this
        ``ImaginaryPowerSpectrumSeries`` instance.
        """
        return '{} ({})'.format(self.__parent.makeLabel(),
                                strings.labels[self])


    def getData(self):
        """Returns the imaginary component of the power spectrum. """

        xdata, ydata = self.__parent.cachedData

        if ydata is not None:
            ydata = ydata.imag

        return xdata, ydata


class MagnitudePowerSpectrumSeries(dataseries.DataSeries):
    """An ``MagnitudePowerSpectrumSeries`` represents the magnitude of a
    complex-valued image. ``MagnitudePowerSpectrumSeries`` instances are
    created by :class:`ComplexPowerSpectrumSeries` instances.
    """


    def __init__(self, parent, *args, **kwargs):
        """Create an ``ImaginaryPowerSpectrumSeries``.

        :arg parent: The :class:`ComplexPowerSpectrumSeries` which owns this
                     ``ImaginaryPowerSpectrumSeries``.

        All other arguments are passed through to the :class:`DataSeries`
        constructor.
        """
        dataseries.DataSeries.__init__(self, *args, **kwargs)
        self.__parent = parent


    def makeLabel(self):
        """Returns a string representation of this
        ``MagnitudePowerSpectrumSeries`` instance.
        """
        return '{} ({})'.format(self.__parent.makeLabel(),
                                strings.labels[self])


    def getData(self):
        """Returns the magnitude of the complex power spectrum. """
        xdata, ydata = self.__parent.cachedData
        if ydata is not None:
            ydata = magnitude(ydata)
        return xdata, ydata


class PhasePowerSpectrumSeries(dataseries.DataSeries):
    """An ``PhasePowerSpectrumSeries`` represents the phase of a complex-valued
    image. ``PhasePowerSpectrumSeries`` instances are created by
    :class:`ComplexPowerSpectrumSeries` instances.
    """


    def __init__(self, parent, *args, **kwargs):
        """Create an ``ImaginaryPowerSpectrumSeries``.

        :arg parent: The :class:`ComplexPowerSpectrumSeries` which owns this
                     ``ImaginaryPowerSpectrumSeries``.

        All other arguments are passed through to the :class:`DataSeries`
        constructor.
        """
        dataseries.DataSeries.__init__(self, *args, **kwargs)
        self.__parent = parent


    def makeLabel(self):
        """Returns a string representation of this ``PhasePowerSpectrumSeries``
        instance.
        """
        return '{} ({})'.format(self.__parent.makeLabel(),
                                strings.labels[self])


    def getData(self):
        """Returns the phase of the complex power spectrum. """
        xdata, ydata = self.__parent.cachedData
        if ydata is not None:
            ydata = phase(ydata)
        return xdata, ydata


class MelodicPowerSpectrumSeries(dataseries.DataSeries,
                                 PowerSpectrumSeries):
    """The ``MelodicPowerSpectrumSeries`` class encapsulates the power spectrum
    of the time course for a single component of a :class:`.MelodicImage`. The
    component is dictated by the :attr:`.NiftiOpts.volume` property.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``MelodicPowerSpectrumSeries``. All arguments are passed
        through to the :meth:`PowerSpectrumSeries.__init__` method.
        """
        dataseries.DataSeries.__init__(self, *args, **kwargs)

        if not isinstance(self.overlay, fslmelimage.MelodicImage):
            raise ValueError('Overlay is not a MelodicImage')

        self.varNorm = False
        self.disableProperty('varNorm')


    def makeLabel(self):
        """Returns a label that can be used for this
        ``MelodicPowerSpectrumSeries``.
        """
        display   = self.displayCtx.getDisplay(self.overlay)
        opts      = display.opts
        component = opts.volume

        return '{} [component {}]'.format(display.name, component + 1)


    def getData(self):
        """Returns the power spectrum for the current component of the
        :class:`.MelodicImage`, as defined by the :attr:`.NiftiOpts.volume`
        property.
        """

        opts      = self.displayCtx.getOpts(self.overlay)
        component = opts.volume

        ydata = self.overlay.getComponentPowerSpectrum(component)
        xdata = np.arange(len(ydata), dtype=np.float32)

        return xdata, ydata


class MeshPowerSpectrumSeries(dataseries.DataSeries,
                              PowerSpectrumSeries):
    """A ``MeshPowerSpectrumSeries`` object encapsulates the power spectrum for
    the data from a :class:`.Mesh` overlay which has some time series
    vertex data associated with it. See the :attr:`.MeshOpts.vertexData`
    property.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``MeshPowerSpectrumSeries`` instance. All arguments are
        passed through to  :meth:`.DataSeries.__init__`.
        """
        dataseries.DataSeries.__init__(self, *args, **kwargs)


    def makeLabel(self):
        """Returns a label to use for this ``MeshPowerSpectrumSeries`` on the
        legend.
        """

        display = self.displayCtx.getDisplay(self.overlay)

        if self.__haveData():
            opts = display.opts
            vidx = opts.getVertex()
            return '{} [{}]'.format(display.name, vidx)

        else:
            return display.name


    def __haveData(self):
        """Returns ``True`` if there is currently time series data to show
        for this ``MeshPowerSpectrumSeries``, ``False`` otherwise.
        """
        opts = self.displayCtx.getOpts(self.overlay)
        vidx = opts.getVertex()
        vd   = opts.getVertexData()

        return vidx is not None and vd is not None and vd.shape[1] > 1


    def getData(self):
        """Returns the power spectrum of the data at the current location for
        the :class:`.Mesh`, or ``None, None`` if there is no data.
        """

        if not self.__haveData():
            return None, None

        opts = self.displayCtx.getOpts(self.overlay)
        vidx = opts.getVertex()

        if vidx is None:
            return None, None

        vd    = opts.getVertexData()
        data  = vd[vidx, :]
        xdata = calcFrequencies(  data, self.sampleTime)
        ydata = calcPowerSpectrum(data, self.varNorm)

        return xdata, ydata



#!/usr/bin/env python
#
# __init__.py - Functions for OpenGL 1.4 rendering.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package contains modules for rendering various :class:`.GLObject`
types in an OpenGL 1.4 compatible manner.
"""

from . import glvolume_funcs      # noqa
from . import glrgbvolume_funcs   # noqa
from . import glrgbvector_funcs   # noqa
from . import gllinevector_funcs  # noqa
from . import glmesh_funcs        # noqa
from . import glmask_funcs        # noqa
from . import gllabel_funcs       # noqa

gltensor_funcs = None
glsh_funcs     = None
glmip_funcs    = None



#!/usr/bin/env python
#
# overlaydisplaytoolbar.py - The OverlayDisplayToolBar.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>

"""This module provides the :class:`OverlyDisplyaToolBar`, a
:class:`.ControlToolBar` containing controls for changing the display settings
of the currently selected overlay.
"""


import            logging
import os.path as op

import wx

import fsleyes_props                  as props
import fsleyes_widgets.utils.typedict as td

import fsleyes.controls.controlpanel  as ctrlpanel
import fsleyes.icons                  as icons
import fsleyes.tooltips               as fsltooltips
import fsleyes.actions                as actions
import fsleyes.colourmaps             as fslcm
import fsleyes.strings                as strings


log = logging.getLogger(__name__)


class OverlayDisplayToolBar(ctrlpanel.ControlToolBar):
    """The ``OverlyDisplyaToolBar`` is a :class:`.ControlToolBar` containing
    controls which allow the user to change the display settings of the
    currently selected overlay (as defined by the
    :attr:`.DisplayContext.selectedOverlay` property). The display settings
    for an overlay are contained in the :class:`.Display` and
    :class:`.DisplayOpts` instances that are associated with that overlay.


    An ``OverlyDisplyaToolBar`` looks something like the following:

    .. image:: images/overlaydisplaytoolbar.png
       :scale: 50%
       :align: center


    The specific controls which are displayed are defined in the
    :attr:`_TOOLBAR_PROPS` dictionary, and are created by the following
    methods:

    .. autosummary::
       :nosignatures:

       _OverlayDisplayToolBar__makeDisplayTools
       _OverlayDisplayToolBar__makeVolumeOptsTools
       _OverlayDisplayToolBar__makeMaskOptsTools
       _OverlayDisplayToolBar__makeLabelOptsTools
       _OverlayDisplayToolBar__makeVectorOptsTools
       _OverlayDisplayToolBar__makeRGBVectorOptsTools
       _OverlayDisplayToolBar__makeLineVectorOptsTools
       _OverlayDisplayToolBar__makeMeshOptsTools
       _OverlayDisplayToolBar__makeGiftiOptsTools
       _OverlayDisplayToolBar__makeFreesurferOptsTools
       _OverlayDisplayToolBar__makeTensorOptsTools
       _OverlayDisplayToolBar__makeSHOptsTools
       _OverlayDisplayToolBar__makeMIPOptsTools
    """

    def __init__(self, parent, overlayList, displayCtx, frame, viewPanel):
        """Create an ``OverlyDisplyaToolBar``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` instance.

        :arg viewPanel:   The :class:`.ViewPanel` which this
                          ``OverlayDisplayToolBar`` is owned by.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          kbFocus=True)

        self.__viewPanel      = viewPanel
        self.__currentOverlay = None

        self.displayCtx.addListener(
            'selectedOverlay',
            self.name,
            self.__selectedOverlayChanged)
        self.overlayList.addListener(
            'overlays',
            self.name,
            self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``OverlyDisplyaToolBar`` is no longer
        needed. Removes some property listeners, and calls the
        :meth:`.ControlToolBar.destroy` method.
        """

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)

        if self.__currentOverlay is not None and \
           self.__currentOverlay in self.overlayList:

            display = self.displayCtx.getDisplay(self.__currentOverlay)
            display.removeListener('overlayType', self.name)
            display.removeListener('enabled',     self.name)

        self.__currentOverlay = None
        self.__viewPanel      = None

        ctrlpanel.ControlToolBar.destroy(self)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OverlayDisplayToolBar`` is only intended to be added to
        :class:`.OrthoPanel`, :class:`.LightBoxPanel`, or
        :class:`.Scene3DPanel` views.
        """
        from fsleyes.views.orthopanel    import OrthoPanel
        from fsleyes.views.lightboxpanel import LightBoxPanel
        from fsleyes.views.scene3dpanel  import Scene3DPanel
        return [OrthoPanel, LightBoxPanel, Scene3DPanel]


    def __showTools(self, overlay):
        """Creates and shows a set of controls allowing the user to change
        the display settings of the specified ``overlay``.
        """

        oldTools = self.GetTools()

        # See long comment at bottom
        def destroyOldTools():
            for t in oldTools:
                t.Destroy()

        for t in oldTools:
            t.Show(False)

        self.ClearTools(destroy=False, postevent=False)

        log.debug('Showing tools for {}'.format(overlay))

        display   = self.displayCtx.getDisplay(overlay)
        opts      = display.opts
        tools     = []
        nav       = []

        # Display tools
        dispTools, dispNav = self.__makeDisplayTools(display)

        tools.extend(dispTools)
        nav  .extend(dispNav)

        # DisplayOpts tools
        makeFunc = getattr(self, '_{}__make{}Tools'.format(
            type(self).__name__, type(opts).__name__), None)

        if makeFunc is not None:
            optsTools, optsNav = makeFunc(opts)

            tools.extend(optsTools)
            nav  .extend(optsNav)

        self.SetTools(   tools)
        self.setNavOrder(nav)

        # This method may have been called via an
        # event handler an existing tool in the
        # toolbar - in this situation, destroying
        # that tool will result in nasty crashes,
        # as the wx widget that generated the event
        # will be destroyed while said event is
        # being processed. So we destroy the old
        # tools asynchronously, well after the event
        # which triggered this method call will have
        # returned.
        wx.CallLater(1000, destroyOldTools)


    def __overlayEnableChanged(self, *a):
        """Called when the :attr:`.Display.enabled` property for the currently
        selected overlay changes. Enables/disables this
        ``OverlayDisplayToolBar`` accordingly.
        """
        display = self.displayCtx.getDisplay(self.__currentOverlay)
        self.Enable(display.enabled)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :class:`.OverlayList` changes. Ensures that controls for the currently
        selected overlay are being shown.
        """

        if self.__currentOverlay is not None and \
           self.__currentOverlay in self.overlayList:
            display = self.displayCtx.getDisplay(self.__currentOverlay)
            display.removeListener('overlayType', self.name)
            display.removeListener('enabled',     self.name)

        overlay = self.displayCtx.getSelectedOverlay()

        self.__currentOverlay = overlay

        if overlay is None:
            self.ClearTools(destroy=True)
            return

        display = self.displayCtx.getDisplay(overlay)

        display.addListener('enabled',
                            self.name,
                            self.__overlayEnableChanged)
        display.addListener('overlayType',
                            self.name,
                            self.__selectedOverlayChanged)

        self.__showTools(overlay)
        self.Enable(display.enabled)


    def __makeDisplayTools(self, display):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.Display` instance.
        """

        viewPanel = self.__viewPanel

        # Display settings
        nameSpec  = _TOOLBAR_PROPS[display, 'name']
        typeSpec  = _TOOLBAR_PROPS[display, 'overlayType']
        alphaSpec = _TOOLBAR_PROPS[display, 'alpha']
        briSpec   = _TOOLBAR_PROPS[display, 'brightness']
        conSpec   = _TOOLBAR_PROPS[display, 'contrast']

        # Buttons which toggle overlay
        # info and display panel
        panelSpec = actions.ToggleActionButton(
            'toggleDisplayPanel',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('gearHighlight24'),
                  icons.findImageFile('gear24')],
            tooltip=fsltooltips.actions[viewPanel, 'toggleDisplayPanel'])

        infoSpec = actions.ToggleActionButton(
            'toggleOverlayInfo',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('informationHighlight24'),
                  icons.findImageFile('information24')],
            tooltip=fsltooltips.actions[viewPanel, 'toggleOverlayInfo'])

        # Name/overlay type and brightness/contrast
        # are respectively placed together
        nameTypePanel = wx.Panel(self)
        briconPanel   = wx.Panel(self)
        nameTypeSizer = wx.BoxSizer(wx.VERTICAL)
        briconSizer   = wx.FlexGridSizer(2, 2, 0, 0)

        briconSizer.AddGrowableCol(1)

        nameTypePanel.SetSizer(nameTypeSizer)
        briconPanel  .SetSizer(briconSizer)

        panelWidget = props.buildGUI(self,          viewPanel, panelSpec)
        infoWidget  = props.buildGUI(self,          viewPanel, infoSpec)
        nameWidget  = props.buildGUI(nameTypePanel, display,   nameSpec)
        typeWidget  = props.buildGUI(nameTypePanel, display,   typeSpec)
        briWidget   = props.buildGUI(briconPanel,   display,   briSpec)
        conWidget   = props.buildGUI(briconPanel,   display,   conSpec)
        alphaWidget = props.buildGUI(self,          display,   alphaSpec)

        briLabel    = wx.StaticText(briconPanel)
        conLabel    = wx.StaticText(briconPanel)

        briLabel.SetLabel(strings.properties[display, 'brightness'])
        conLabel.SetLabel(strings.properties[display, 'contrast'])

        # name/type panel
        nameTypeSizer.Add(nameWidget, flag=wx.EXPAND)
        nameTypeSizer.Add(typeWidget, flag=wx.EXPAND)

        # opacity is given a label
        alphaPanel = self.MakeLabelledTool(
            alphaWidget, strings.properties[display, 'alpha'])

        # bricon panel
        briconSizer.Add(briLabel)
        briconSizer.Add(briWidget)
        briconSizer.Add(conLabel)
        briconSizer.Add(conWidget)

        tools = [panelWidget,
                 infoWidget,
                 nameTypePanel,
                 alphaPanel,
                 briconPanel]
        nav   = [panelWidget,
                 infoWidget,
                 nameWidget,
                 typeWidget,
                 alphaWidget,
                 briWidget,
                 conWidget]

        return tools, nav


    def __makeVolumeOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.VolumeOpts` instance.
        """
        rangeSpec      = _TOOLBAR_PROPS[opts, 'displayRange']
        resetSpec      = _TOOLBAR_PROPS[opts, 'resetDisplayRange']
        cmapSpec       = _TOOLBAR_PROPS[opts, 'cmap']
        negCmapSpec    = _TOOLBAR_PROPS[opts, 'negativeCmap']
        useNegCmapSpec = _TOOLBAR_PROPS[opts, 'useNegativeCmap']

        cmapPanel = wx.Panel(self)

        rangeWidget      = props.buildGUI(self,      opts, rangeSpec)
        resetWidget      = props.buildGUI(self,      opts, resetSpec)
        useNegCmapWidget = props.buildGUI(self,      opts, useNegCmapSpec)
        cmapWidget       = props.buildGUI(cmapPanel, opts, cmapSpec)
        negCmapWidget    = props.buildGUI(cmapPanel, opts, negCmapSpec)

        cmapSizer = wx.BoxSizer(wx.VERTICAL)
        cmapPanel.SetSizer(cmapSizer)
        cmapSizer.Add(cmapWidget)
        cmapSizer.Add(negCmapWidget)

        tools = [resetWidget, rangeWidget, useNegCmapWidget, cmapPanel]
        nav   = [resetWidget, rangeWidget, useNegCmapWidget, cmapWidget,
                 negCmapWidget]

        return tools, nav


    def __makeComplexOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.ComplexOpts` instance.
        """
        rangeSpec = _TOOLBAR_PROPS[opts, 'displayRange']
        resetSpec = _TOOLBAR_PROPS[opts, 'resetDisplayRange']
        cmapSpec  = _TOOLBAR_PROPS[opts, 'cmap']
        compSpec  = _TOOLBAR_PROPS[opts, 'component']
        ccpanel   = wx.Panel(self)

        rangeWidget = props.buildGUI(self,    opts, rangeSpec)
        resetWidget = props.buildGUI(self,    opts, resetSpec)
        cmapWidget  = props.buildGUI(ccpanel, opts, cmapSpec)
        compWidget  = props.buildGUI(ccpanel, opts, compSpec)

        ccsizer = wx.BoxSizer(wx.VERTICAL)
        ccpanel.SetSizer(ccsizer)
        ccsizer.Add(cmapWidget, flag=wx.EXPAND)
        ccsizer.Add(compWidget, flag=wx.EXPAND)

        tools = [resetWidget, rangeWidget, ccpanel]
        nav   = [resetWidget, rangeWidget, cmapWidget, compWidget]

        return tools, nav


    def __makeMaskOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.MaskOpts` instance.
        """
        thresSpec  = _TOOLBAR_PROPS[opts, 'threshold']
        colourSpec = _TOOLBAR_PROPS[opts, 'colour']

        thresWidget  = props.buildGUI(self, opts, thresSpec)
        colourWidget = props.buildGUI(self, opts, colourSpec)

        tools = [thresWidget, colourWidget]

        return tools, tools


    def __makeLabelOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.LabelOpts` instance.
        """

        lutSpec     = _TOOLBAR_PROPS[opts, 'lut']
        outlineSpec = _TOOLBAR_PROPS[opts, 'outline']
        widthSpec   = _TOOLBAR_PROPS[opts, 'outlineWidth']

        # lut/outline width widgets
        # are on a single panel
        lutWidthPanel = wx.Panel(self)
        lutWidthSizer = wx.FlexGridSizer(2, 2, 0, 0)
        lutWidthPanel.SetSizer(lutWidthSizer)

        lutWidget     = props.buildGUI(lutWidthPanel, opts, lutSpec)
        widthWidget   = props.buildGUI(lutWidthPanel, opts, widthSpec)
        outlineWidget = props.buildGUI(self,          opts, outlineSpec)

        lutLabel   = wx.StaticText(lutWidthPanel)
        widthLabel = wx.StaticText(lutWidthPanel)

        lutLabel  .SetLabel(strings.properties[opts, 'lut'])
        widthLabel.SetLabel(strings.properties[opts, 'outlineWidth'])

        lutWidthSizer.Add(lutLabel)
        lutWidthSizer.Add(lutWidget,   flag=wx.EXPAND)
        lutWidthSizer.Add(widthLabel)
        lutWidthSizer.Add(widthWidget, flag=wx.EXPAND)

        tools = [lutWidthPanel, outlineWidget]
        nav   = [lutWidget, widthWidget, outlineWidget]

        return tools, nav


    def __makeVectorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.VectorOpts` instance.
        """

        modSpec   = _TOOLBAR_PROPS[opts, 'modulateImage']
        clipSpec  = _TOOLBAR_PROPS[opts, 'clipImage']
        rangeSpec = _TOOLBAR_PROPS[opts, 'clippingRange']

        panel = wx.Panel(self)
        sizer = wx.GridBagSizer()
        panel.SetSizer(sizer)

        modWidget   = props.buildGUI(panel, opts, modSpec)
        clipWidget  = props.buildGUI(panel, opts, clipSpec)
        rangeWidget = props.buildGUI(panel, opts, rangeSpec)
        modLabel    = wx.StaticText(panel)
        clipLabel   = wx.StaticText(panel)

        modLabel .SetLabel(strings.properties[opts, 'modulateImage'])
        clipLabel.SetLabel(strings.properties[opts, 'clipImage'])

        sizer.Add(modLabel,    pos=(0, 0), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(modWidget,   pos=(0, 1), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(clipLabel,   pos=(1, 0), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(clipWidget,  pos=(1, 1), span=(1, 1), flag=wx.EXPAND)
        sizer.Add(rangeWidget, pos=(0, 2), span=(2, 1), flag=wx.EXPAND)

        tools = [panel]
        nav   = [modWidget, clipWidget, rangeWidget]

        return tools, nav


    def __makeRGBVectorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.RGBVectorOpts` instance.
        """
        return self.__makeVectorOptsTools(opts)


    def __makeLineVectorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.LineVectorOpts` instance.
        """
        widthSpec = _TOOLBAR_PROPS[opts, 'lineWidth']

        widget    = props.buildGUI(self, opts, widthSpec)
        lblWidget = self.MakeLabelledTool(
            widget, strings.properties[opts, 'lineWidth'])

        tools, nav = self.__makeVectorOptsTools(opts)

        return tools + [lblWidget], nav + [widget]


    def __makeMeshOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.MeshOpts` instance.
        """
        colourSpec  = _TOOLBAR_PROPS[opts, 'colour']
        outlineSpec = _TOOLBAR_PROPS[opts, 'outline']
        widthSpec   = _TOOLBAR_PROPS[opts, 'outlineWidth']

        colourWidget  = props.buildGUI(self, opts, colourSpec)
        outlineWidget = props.buildGUI(self, opts, outlineSpec)
        widthWidget   = props.buildGUI(self, opts, widthSpec)

        lblWidthWidget  = self.MakeLabelledTool(
            widthWidget, strings.properties[opts, 'outlineWidth'])

        tools = [colourWidget, outlineWidget, lblWidthWidget]
        nav   = [colourWidget, outlineWidget, widthWidget]

        return tools, nav


    def __makeGiftiOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.GiftiOpts` instance.
        """
        tools, nav = self.__makeMeshOptsTools(opts)

        vertWidget  = _TOOLBAR_PROPS[opts, 'vertexSet']
        vdataWidget = _TOOLBAR_PROPS[opts, 'vertexData']

        panel       = wx.Panel(self)
        sizer       = wx.BoxSizer(wx.VERTICAL)
        vertWidget  = props.buildGUI(panel, opts, vertWidget)
        vdataWidget = props.buildGUI(panel, opts, vdataWidget)

        sizer.Add(vertWidget,  flag=wx.EXPAND)
        sizer.Add(vdataWidget, flag=wx.EXPAND)
        panel.SetSizer(sizer)

        tools += [panel]
        nav   += [vertWidget, vdataWidget]

        return tools, nav


    def __makeFreesurferOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.FreesurferOpts` instance.
        """
        return self.__makeGiftiOptsTools(opts)


    def __makeTensorOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.TensorOpts` instance.
        """
        lightingSpec   = _TOOLBAR_PROPS[opts, 'lighting']
        lightingWidget = props.buildGUI(self, opts, lightingSpec)

        tools, nav = self.__makeVectorOptsTools(opts)

        return tools + [lightingWidget], nav + [lightingWidget]


    def __makeSHOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.SHOpts` instance.
        """
        sizeSpec = _TOOLBAR_PROPS[opts, 'size']
        radSpec  = _TOOLBAR_PROPS[opts, 'radiusThreshold']

        panel = wx.Panel(self)
        sizer = wx.FlexGridSizer(2, 2, 0, 0)

        sizeWidget  = props.buildGUI(panel, opts, sizeSpec)
        radWidget   = props.buildGUI(panel, opts, radSpec)

        sizeLabel = wx.StaticText(panel)
        radLabel  = wx.StaticText(panel)

        sizeLabel.SetLabel(strings.properties[opts, 'size'])
        radLabel .SetLabel(strings.properties[opts, 'radiusThreshold'])

        sizer.Add(sizeLabel)
        sizer.Add(sizeWidget)
        sizer.Add(radLabel)
        sizer.Add(radWidget)
        panel.SetSizer(sizer)

        tools, nav = self.__makeVectorOptsTools(opts)

        tools.extend([panel])
        nav.extend(  [sizeWidget, radWidget])

        return tools, nav


    def __makeMIPOptsTools(self, opts):
        """Creates and returns a collection of controls for editing properties
        of the given :class:`.MIPOpts` instance.
        """
        rangeSpec = _TOOLBAR_PROPS[opts, 'displayRange']
        resetSpec = _TOOLBAR_PROPS[opts, 'resetDisplayRange']
        cmapSpec  = _TOOLBAR_PROPS[opts, 'cmap']

        cmapPanel = wx.Panel(self)

        rangeWidget = props.buildGUI(self,      opts, rangeSpec)
        resetWidget = props.buildGUI(self,      opts, resetSpec)
        cmapWidget  = props.buildGUI(cmapPanel, opts, cmapSpec)

        cmapSizer = wx.BoxSizer(wx.VERTICAL)
        cmapPanel.SetSizer(cmapSizer)
        cmapSizer.Add(cmapWidget)

        tools = [resetWidget, rangeWidget, cmapPanel]
        nav   = [resetWidget, rangeWidget, cmapWidget]

        return tools, nav


def _imageLabel(img):
    """Used to generate labels for the :attr:`.VectorOpts.modulateImage`,
    :attr:`.VectorOpts.clipImage`, and other :class:`.Image`-based
    choice properties.
    """
    if img is None: return 'None'
    else:           return img.name



def _pathLabel(p):
    if p is None: return 'None'
    else:         return op.basename(p)


_TOOLTIPS = td.TypeDict({

    'Display.name'        : fsltooltips.properties['Display.name'],
    'Display.overlayType' : fsltooltips.properties['Display.overlayType'],
    'Display.alpha'       : fsltooltips.properties['Display.alpha'],
    'Display.brightness'  : fsltooltips.properties['Display.brightness'],
    'Display.contrast'    : fsltooltips.properties['Display.contrast'],

    'ColourMapOpts.displayRange' :
    fsltooltips.properties['ColourMapOpts.displayRange'],
    'VolumeOpts.cmap'              :
    fsltooltips.properties['ColourMapOpts.cmap'],
    'VolumeOpts.negativeCmap'      :
    fsltooltips.properties['ColourMapOpts.negativeCmap'],
    'VolumeOpts.useNegativeCmap'   :
    fsltooltips.properties['ColourMapOpts.useNegativeCmap'],

    'ComplexOpts.component' : fsltooltips.properties['ComplexOpts.component'],

    'MaskOpts.threshold' : fsltooltips.properties['MaskOpts.threshold'],
    'MaskOpts.colour'    : fsltooltips.properties['MaskOpts.colour'],

    'LabelOpts.lut'          : fsltooltips.properties['LabelOpts.lut'],

    'LabelOpts.outline'      : fsltooltips.properties['LabelOpts.outline'],
    'LabelOpts.outlineWidth' : fsltooltips.properties['LabelOpts.'
                                                      'outlineWidth'],

    'VectorOpts.modulateImage' : fsltooltips.properties['VectorOpts.'
                                                        'modulateImage'],
    'VectorOpts.clipImage'     : fsltooltips.properties['VectorOpts.'
                                                        'clipImage'],
    'VectorOpts.clippingRange' : fsltooltips.properties['VectorOpts.'
                                                        'clippingRange'],

    'LineVectorOpts.lineWidth' : fsltooltips.properties['LineVectorOpts.'
                                                        'lineWidth'],

    'MeshOpts.colour'       : fsltooltips.properties['MeshOpts.colour'],
    'MeshOpts.outline'      : fsltooltips.properties['MeshOpts.outline'],
    'MeshOpts.outlineWidth' : fsltooltips.properties['MeshOpts.'
                                                     'outlineWidth'],
    'MeshOpts.vertexSet'    : fsltooltips.properties['MeshOpts.vertexSet'],
    'MeshOpts.vertexData'   : fsltooltips.properties['MeshOpts.vertexData'],

    'TensorOpts.lighting' : fsltooltips.properties['TensorOpts.'
                                                   'lighting'],

    'SHOpts.lighting'        : fsltooltips.properties['SHOpts.lighting'],
    'SHOpts.size'            : fsltooltips.properties['SHOpts.size'],
    'SHOpts.radiusThreshold' : fsltooltips.properties['SHOpts.'
                                                      'radiusThreshold'],
})
"""This dictionary contains tooltips for :class:`.Display` and
:class:`.DisplayOpts` properties. It is referenced in the
:attr:`_TOOLBAR_PROPS` dictionary definition.
"""


_TOOLBAR_PROPS = td.TypeDict({

    'Display.name'         : props.Widget(
        'name',
        tooltip=_TOOLTIPS['Display.name']),
    'Display.overlayType'  : props.Widget(
        'overlayType',
        tooltip=_TOOLTIPS['Display.overlayType'],
        labels=strings.choices['Display.overlayType']),
    'Display.alpha'        : props.Widget(
        'alpha',
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['Display.alpha']),
    'Display.brightness'   : props.Widget(
        'brightness',
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['Display.brightness']),
    'Display.contrast'     : props.Widget(
        'contrast',
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['Display.contrast']),

    'VolumeOpts.resetDisplayRange' : actions.ActionButton(
        'resetDisplayRange',
        icon=icons.findImageFile('verticalReset32')
    ),

    'VolumeOpts.displayRange' : props.Widget(
        'displayRange',
        slider=False,
        showLimits=False,
        spinWidth=10,
        tooltip=_TOOLTIPS['ColourMapOpts.displayRange'],
        labels=[strings.choices['ColourMapOpts.displayRange.min'],
                strings.choices['ColourMapOpts.displayRange.max']]),
    'VolumeOpts.cmap' : props.Widget(
        'cmap',
        labels=fslcm.getColourMapLabel,
        tooltip=_TOOLTIPS['VolumeOpts.cmap']),
    'VolumeOpts.useNegativeCmap' : props.Widget(
        'useNegativeCmap',
        icon=[icons.findImageFile('twocmaps24'),
              icons.findImageFile('onecmap24')],
        toggle=True,
        tooltip=_TOOLTIPS['VolumeOpts.useNegativeCmap']),

    'VolumeOpts.negativeCmap' : props.Widget(
        'negativeCmap',
        labels=fslcm.getColourMapLabel,
        tooltip=_TOOLTIPS['VolumeOpts.negativeCmap'],
        dependencies=['useNegativeCmap'],
        enabledWhen=lambda i, unc : unc),

    'ComplexOpts.component' : props.Widget(
        'component',
        labels=strings.choices['ComplexOpts.component'],
        tooltip=_TOOLTIPS['ComplexOpts.component']),

    'MaskOpts.threshold' : props.Widget(
        'threshold',
        showLimits=False,
        spin=True,
        tooltip=_TOOLTIPS['MaskOpts.threshold'],
        labels=[strings.choices['MaskOpts.threshold.min'],
                strings.choices['MaskOpts.threshold.max']]),
    'MaskOpts.colour'    : props.Widget(
        'colour',
        size=(24, 24),
        tooltip=_TOOLTIPS['MaskOpts.colour']),

    'LabelOpts.lut'     : props.Widget(
        'lut',
        tooltip=_TOOLTIPS['LabelOpts.lut'],
        labels=lambda l: l.name),
    'LabelOpts.outline' : props.Widget(
        'outline',
        tooltip=_TOOLTIPS['LabelOpts.outline'],
        icon=[icons.findImageFile('outline24'),
              icons.findImageFile('filled24')],
        toggle=True),
    'LabelOpts.outlineWidth' : props.Widget(
        'outlineWidth',
        tooltip=_TOOLTIPS['LabelOpts.outlineWidth'],
        showLimits=False,
        spin=False),

    'MeshOpts.colour'       : props.Widget(
        'colour',
        size=(24, 24),
        tooltip=_TOOLTIPS['MeshOpts.colour']),
    'MeshOpts.outline'      : props.Widget(
        'outline',
        tooltip=_TOOLTIPS['MeshOpts.outline'],
        icon=[icons.findImageFile('outline24'),
              icons.findImageFile('filled24')],
        toggle=True),
    'MeshOpts.outlineWidth' : props.Widget(
        'outlineWidth',
        showLimits=False,
        spin=False,
        tooltip=_TOOLTIPS['MeshOpts.outlineWidth'],
        enabledWhen=lambda i: i.outline),
    'MeshOpts.vertexSet' : props.Widget(
        'vertexSet',
        labels=_pathLabel,
        tooltip=_TOOLTIPS['MeshOpts.vertexSet']
    ),
    'MeshOpts.vertexData' : props.Widget(
        'vertexData',
        labels=_pathLabel,
        tooltip=_TOOLTIPS['MeshOpts.vertexData']
    ),

    'VectorOpts.modulateImage' : props.Widget(
        'modulateImage',
        labels=_imageLabel,
        tooltip=_TOOLTIPS['VectorOpts.modulateImage']),
    'VectorOpts.clipImage' : props.Widget(
        'clipImage',
        labels=_imageLabel,
        tooltip=_TOOLTIPS['VectorOpts.clipImage']),
    'VectorOpts.clippingRange' : props.Widget(
        'clippingRange',
        showLimits=False,
        slider=True,
        spin=False,
        tooltip=_TOOLTIPS['VectorOpts.clippingRange'],
        labels=[strings.choices['VectorOpts.clippingRange.min'],
                strings.choices['VectorOpts.clippingRange.max']],
        dependencies=['clipImage'],
        enabledWhen=lambda o, ci: ci is not None),

    'LineVectorOpts.lineWidth' : props.Widget(
        'lineWidth',
        showLimits=False,
        spin=False,
        tooltip=_TOOLTIPS['LineVectorOpts.lineWidth']),

    'TensorOpts.lighting'      : props.Widget(
        'lighting',
        icon=[icons.findImageFile('lightbulbHighlight24'),
              icons.findImageFile('lightbulb24')],
        tooltip=_TOOLTIPS['TensorOpts.lighting']),

    'SHOpts.lighting'        : props.Widget(
        'lighting',
        icon=[icons.findImageFile('lightbulbHighlight24'),
              icons.findImageFile('lightbulb24')],
        tooltip=_TOOLTIPS['SHOpts.lighting']),
    'SHOpts.size'            : props.Widget(
        'size',
        showLimits=False,
        slider=True,
        spin=False,
        tooltip=_TOOLTIPS['SHOpts.size']),
    'SHOpts.radiusThreshold' : props.Widget(
        'radiusThreshold',
        slider=True,
        spin=False,
        showLimits=False,
        tooltip=_TOOLTIPS['SHOpts.radiusThreshold']),


    'MIPOpts.resetDisplayRange' : actions.ActionButton(
        'resetDisplayRange',
        icon=icons.findImageFile('verticalReset32')
    ),

    'MIPOpts.displayRange' : props.Widget(
        'displayRange',
        slider=False,
        showLimits=False,
        spinWidth=10,
        tooltip=_TOOLTIPS['ColourMapOpts.displayRange'],
        labels=[strings.choices['ColourMapOpts.displayRange.min'],
                strings.choices['ColourMapOpts.displayRange.max']]),
    'MIPOpts.cmap' : props.Widget(
        'cmap',
        labels=fslcm.getColourMapLabel,
        tooltip=_TOOLTIPS['VolumeOpts.cmap']),
})
"""This dictionary defines specifications for all controls shown on an
:class:`OverlayDisplayToolBar`.
"""
